from __future__ import annotations
import os
import json
import time
import uuid
import tempfile
from urllib.parse import urlparse
from typing import Optional, Union, Any
import pandas as pd
from arcgis.gis import GIS, Item
from requests.utils import quote
import xml.etree.ElementTree as ET
from .exceptions import ServerError
import requests

########################################################################


class SurveyManager:
    """
    Survey Manager allows users and administrators of Survey 123 to
    analyze, report on, and access the data for various surveys.

    """

    _baseurl = None
    _gis = None
    _portal = None
    _url = None
    _properties = None
    # ----------------------------------------------------------------------

    def __init__(self, gis, baseurl=None):
        """Constructor"""
        if baseurl is None:
            baseurl = "survey123.arcgis.com"
        self._baseurl = baseurl
        self._gis = gis

    # ----------------------------------------------------------------------
    def __str__(self):
        return "< SurveyManager @ {iid} >".format(iid=self._gis._url)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()

    # ----------------------------------------------------------------------
    @property
    def surveys(self) -> list:
        """returns a list of existing Survey"""
        query = (
            'type:"Form" AND NOT tags:"noxlsform"'
            'AND NOT tags:"draft" AND NOT typekeyw'
            "ords:draft AND owner:{owner}"
        ).format(owner=self._gis.users.me.username)
        content = self._gis.content
        items = content.search(
            query=query,
            item_type=None,
            sort_field="avgRating",
            sort_order="desc",
            max_items=10000,
            outside_org=False,
            categories=None,
            category_filters=None,
        )
        return [Survey(item=i, sm=self) for i in items]

    # ----------------------------------------------------------------------
    def get(self, survey_id: Union[Item, str]):
        """returns a single :class:`~arcgis.apps.survey123.Survey` object from and Item ID or Item"""
        if isinstance(survey_id, Item):
            survey_id = survey_id.id
        item = self._gis.content.get(survey_id)
        return Survey(item=item, sm=self)

    # ----------------------------------------------------------------------
    def _xform2webform(xform, portalUrl, connectVersion=None):
        """Converts a XForm XML to Enketo Web form by Enketo Transformer"""
        (dir_path, file_name) = os.path.split(xform)
        xlsx_name = os.path.splitext(file_name)[0]

        # xform_tree = ET.parse(xform)
        # root = xform_tree.getroot()
        # xform_string = ET.tostring(root, encoding='utf8', method='xml')

        with open(xform, "r", encoding="utf-8") as intext:
            xform_string = intext.read()

        url = "https://survey123.arcgis.com/api/xform2webform"
        params = {"xform": xform_string}
        if connectVersion:
            params["connectVersion"] = connectVersion
        try:
            r = requests.post(url, params)
            response_json = r.json()
            r.close()
        except requests.exceptions.ConnectionError as c:
            return "Unable to complete request with message: " + str(c)
        except requests.exceptions.Timeout as t:
            return "Connection timed out: " + str(t)

        else:
            with open(
                os.path.join(dir_path, xlsx_name + ".webform"), "w", encoding="utf-8"
            ) as fp:
                # with open(os.path.join(dir_path, xlsx_name + ".webform"), 'w') as fp:
                response_json["surveyFormJson"]["portalUrl"] = portalUrl
                webform = {
                    "form": response_json["form"],
                    "languageMap": response_json["languageMap"],
                    "model": response_json["model"],
                    "success": response_json["success"],
                    "surveyFormJson": response_json["surveyFormJson"],
                    "transformerVersion": response_json["transformerVersion"],
                }

                fp.write(json.dumps(webform, indent=2))
                # fp.write(json.dumps(response_json, indent=2))
                # fp.close()
            return os.path.join(dir_path, xlsx_name + ".webform")

    # ----------------------------------------------------------------------
    def _xls2xform(self, file_path: str):
        """
        Converts a XLSForm spreadsheet to XForm XML. The spreadsheet must be in Excel XLS(X) format

        ============   ================================================
        *Inputs*       *Description*
        ------------   ------------------------------------------------
        file_path      Required String. Path to the XLS(X) file.
        ============   ================================================

        :returns: dict

        """

        url = "https://{base}/api/xls2xform".format(base=self._baseurl)
        params = {"f": "json"}
        file = {"xlsform": file_path}
        isinstance(self._gis, GIS)
        return self._gis._con.post(
            path=url, postdata=params, files=file, verify_cert=False
        )

    # ----------------------------------------------------------------------
    def _create(
        self,
        project_name: str,
        survey_item: Item,
        summary: str = None,
        tags: str = None,
    ) -> bool:
        """TODO: implement create survery from xls"""
        # XLS Item or File Path
        # https://survey123.arcgis.com/api/xls2xform
        ##Content-Disposition: form-data; name="xlsform"; filename="Form_2.xlsx"
        ##Content-Type: application/octet-stream
        # Create Folder
        # Create Feature Service
        # Update Feature layer and tables
        # Enable editor tracking
        # Update capabilities
        # Create web form
        # Create form item
        # Refresh ?
        return


########################################################################
class Survey:
    """
    A `Survey` is a single instance of a survey project. This class contains
    the :class:`~arcgis.gis.Item` information and properties to access the underlying dataset
    that was generated by the `Survey` form.

    Data can be exported to `Pandas DataFrames`, `shapefiles`, `CSV`, and
    `File Geodatabases`.

    In addition to exporting data to various formats, a `Survey's` data can
    be exported as reports.

    """

    _gis = None
    _sm = None
    _si = None
    _ssi = None
    _baseurl = None
    # ----------------------------------------------------------------------

    def __init__(self, item, sm, baseurl: Optional[str] = None):
        """Constructor"""
        if baseurl is None:
            baseurl = "survey123.arcgis.com"
        self._si = item
        self._gis = item._gis
        self._sm = sm
        try:
            self.layer_name = self._find_layer_name()
        except:
            self.layer_name = None
        self._baseurl = baseurl

        sd = self._si.related_items("Survey2Data", direction="forward")
        if len(sd) > 0:
            for item in sd:
                if "StakeholderView" in item.typeKeywords:
                    self._stk = item
                    _stk_layers = self._stk.layers + self._stk.tables
                    _idx = 0
                    if self.layer_name:
                        for layer in _stk_layers:
                            if layer.properties["name"] == self.layer_name:
                                _idx = layer.properties["id"]
                    self._stk_url = self._stk.url + f"/{str(_idx)}"

        related = self._si.related_items("Survey2Service", direction="forward")
        if len(related) > 0:
            self._ssi = related[0]
            self._ssi_layers = self._ssi.layers + self._ssi.tables

            ssi_layer = None
            if self.layer_name:
                for layer in self._ssi_layers:
                    if layer.properties["name"] == self.layer_name:
                        _idx = layer.properties["id"]
                        ssi_layer = layer
                        break
            if not ssi_layer:
                ssi_layer = self._ssi_layers[0]
                _idx = ssi_layer.properties["id"]
            self._ssi_url = ssi_layer._url
            try:
                if self._ssi_layers[0].properties["isView"] == True:
                    view_url = ssi_layer._url[:-1]
                    self.parent_fl_url = self._find_parent(view_url) + f"/{str(_idx)}"
            except KeyError:
                self.parent_fl_url = ssi_layer._url

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the properties of the survey"""
        return dict(self._si)

    # ----------------------------------------------------------------------
    def __str__(self):
        return "<Survey @ {iid}>".format(iid=self._si.title)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()

    # ----------------------------------------------------------------------
    def download(
        self, export_format: str, save_folder: Optional[str] = None
    ) -> Union[str, pd.Dataframe]:
        """
        Exports the Survey's data to other format

        ================  ===============================================================
        **Parameter**      **Description**
        ----------------  ---------------------------------------------------------------
        export_format     Required String. This is the acceptable export format that a
                          user can export the survey data to. The following formats are
                          acceptable: File Geodatabase, Shapefile, CSV, and DF.
        ----------------  ---------------------------------------------------------------
        save_folder       Optional String. Specify the folder location where the output file should be stored.
        ================  ===============================================================

        :Returns: String or DataFrame
        """

        title = "a%s" % uuid.uuid4().hex
        if export_format.lower() == "df":
            return self._ssi.layers[0].query().sdf
        if save_folder is None:
            save_folder = tempfile.gettempdir()
        isinstance(self._ssi, Item)
        eitem = self._ssi.export(
            title=title,
            export_format=export_format,
        )
        save_file = eitem.download(save_path=save_folder)
        eitem.delete(force=True)
        return save_file

    # ----------------------------------------------------------------------
    def generate_report(
        self,
        report_template: Item,
        where: str = "1=1",
        utc_offset: str = "+00:00",
        report_title: Optional[str] = None,
        package_name: Optional[str] = None,
        output_format: str = "docx",
        folder_id: Optional[str] = None,
        merge_files: Optional[str] = None,
        survey_item: Optional[Item] = None,
        webmap_item: Optional[Item] = None,
        map_scale: Optional[float] = None,
        locale: str = "en",
        save_folder: Optional[str] = None,
    ) -> str:
        """
        The `generate_report` method allows users to create Microsoft Word and PDF reports
        from a survey using a report template. Reports are saved as an :class:`~arcgis.gis.Item` in an ArcGIS
        content folder or saved locally on disk. For additional information on parameters,
        see `Create Report <https://developers.arcgis.com/survey123/api-reference/rest/report/#create-report>`.

        .. note::
            The Survey123 report service may output one or more `.docx` or `.pdf` files, or a zipped
            package of these files. Whether the output is contained in a `.zip` file depends
            on the number of files generated and their size. For more information, see the
            `packageFiles` parameter in the `Create Report <https://developers.arcgis.com/survey123/api-reference/rest/report/#request-parameters-3>`_ documentation.

        .. note::
            To save to disk, do not specify a `folder_id` argument.

        ================  ===============================================================
        **Parameter**      **Description**
        ----------------  ---------------------------------------------------------------
        report_template   Required :class:`~arcgis.gis.Item`. The report template.
        ----------------  ---------------------------------------------------------------
        where             Optional String. The select statement issued on survey
                          :class:`~arcgis.features.FeatureLayer` to report on
                          all survey records or a subset.

                          Query the `parent_fl_url` property of the
                          :class:`~arcgis.apps.survey123.Survey` object to get the
                          feature layer URL and retrieve a list of fields.

                          .. code-block:: python

                              >>> gis = GIS(profile="your_profile")
                              >>> smgr = SurveyManager(gis)

                              >>> survey_item = gis.content.get("<survey form id>")
                              >>> survey_obj = smgr.get(survey_item.id)

                              >>> survey_fl = FeatureLayer(survey_obj.parent_fl_url, gis)

                              >>> print([f["name"] for f in survey_fl.properties.fields])
        ----------------  ---------------------------------------------------------------
        utc_offset        Optional String.  Time offset from UTC. This offset is applied to
                          all `date`, `time`, and `dateTime` questions that appear in the report output.
                          Example: EST - "+04:00"
        ----------------  ---------------------------------------------------------------
        report_title      Optional String. If `folder_id` is provided, the result is an
                          :class:`~arcgis.gis.Item` with this argument as the title. If
                          `save_folder` argument is provided, this argument will be the
                          name of the output file, or the base name for files
                          in the output zipped package if the server-side component
                          chose to zip up the output (depends upon the size and number
                          of files that would result).


                          .. note::
                              If `merge_files` is either `nextPage` or `continuous`,
                              `report_title` is the output file name.
        ----------------  ---------------------------------------------------------------
        package_name      Optional String. Specify the file name (without extension) of the
                          packaged `.zip` file. If multiple files are packaged, the `report_title`
                          argument will be used to name individual files in the package.


                          .. note::
                            The Survey123 report service automatically decides whether to package
                            generated reports as a `.zip` file, depending on the output file count.
                            See the `packageFiles` parameter description in the `Create Report Request parameters <https://developers.arcgis.com/survey123/api-reference/rest/report/#request-parameters-3>`_
                            documentation for details.
        ----------------  ---------------------------------------------------------------
        save_folder       Optional String. Specify the folder location where the output
                          file or zipped file should be stored. If `folder_id` argument
                          is provided, this argument is ignored.
        ----------------  ---------------------------------------------------------------
        output_format     Optional String. Accepts `docx` or `pdf`.
        ----------------  ---------------------------------------------------------------
        folder_id         Optional String. If a file :class:`~arcgis.gis.Item` is the
                          desired output, specify the `id` value of the ArcGIS content
                          folder.
        ----------------  ---------------------------------------------------------------
        merge_files       Optional String. Specify if output is a single file containing individual
                          records on multiple pages (`nextPage` or `continuous`) or
                          multiple files (`none`).

                          + `none` - Print multiple records in split mode. Each record
                            is a separate file. This is the default value.
                          + `nextPage` - Print multiple records in a single document.
                            Each record starts on a new page.
                          + `continuous` - Print multiple records in a single document.
                            EAch records starts on the same page of the previous record.

                          .. note::
                              A merged file larger than 500 MB will be split into multiple
                              files.
        ----------------  ---------------------------------------------------------------
        survey_item       Optional survey :class:`~arcgis.gis.Item` to provide
                          additional information on survey structure.
        ----------------  ---------------------------------------------------------------
        webmap_item       Optional web map :class:`~arcgis.gis.Item`. Specify the basemap for all
                          map questions in the report. This takes precedence over the map set for
                          each question in the report template.
        ----------------  ---------------------------------------------------------------
        map_scale         Optional Float. Specify the map scale for all map questions in the report.
                          The map will center on the feature geometry. This takes precedence over the
                          scale set for each question in the report template.
        ----------------  ---------------------------------------------------------------
        locale            Optional String. Specify the locale to format number
                          and date values.
        ================  ===============================================================

        :Returns:
            An :class:`~arcgis.gis.Item` or string upon completion of the reporting
            `job <https://developers.arcgis.com/survey123/api-reference/rest/report/#jobs>`_.
            For details on the returned value, see `Response Parameters <https://developers.arcgis.com/survey123/api-reference/rest/report/#response-parameters>`_
            for the :func:`~arcgis.apps.survey123.Survey.generate_report` job.

        .. code-block:: python

            # Usage example #1: output a PDF file Item:
            >>> from arcgis.gis import GIS
            >>> from arcgis.apps.survey123 import SurveyManager

            >>> gis = GIS(profile="your_profile_name")

            >>> # Get report template and survey items
            >>> report_templ = gis.content.get("<template item id>")
            >>> svy_item = gis.content.get("<survey item id>")

            >>> svy_mgr = SurveyManager(gis)
            >>> svy_obj = svy_mgr.get(svy_item.id)

            >>> user_folder_id = [f["id"]
                                 for f in gis.users.me.folders
                                 if f["title"] == "folder_title"][0]

            >>> report_item = svy_obj.generate_report(report_template=report_templ,
                                                      report_title="Title of Report Item",
                                                      output_format="pdf",
                                                      folder_id=user_folder_id,
                                                      merge_files="continuous")

           # Usage example #2: output a Microsoft Word document named `LessThan20_Report.docx`

           >>> report_file = svy_obj.generate_report(report_template=report_templ,
                                                     where="objectid < 20",
                                                     report_title="LessThan20_Report",
                                                     output_format="docx",
                                                     save_folder="file\system\directory\",
                                                     merge_files="nextPage")

           # Usage example #3: output a zip file named `api_gen_report_pkg.zip` of individual
           #                   pdf files with a base name of `SpecimensOver30`

           >>> report_file = svy_obj.generate_report(report_template=report_templ,
                                                     where="number_specimens>30",
                                                     report_title="SpecimensOver30",
                                                     output_format="pdf",
                                                     save_folder="file\system\directory",
                                                     package_name="api_gen_report_pkg")

        """
        if isinstance(where, str):
            where = {"where": where}

        url = "https://{base}/api/featureReport/createReport/submitJob".format(
            base=self._baseurl
        )

        try:
            if (
                self._si._gis.users.me.username == self._si.owner
                and self._ssi_layers[0].properties["isView"] == True
            ):
                fl_url = self.parent_fl_url
            elif self._si._gis.users.me.username != self._si.owner:
                fl_url = self._stk_url
        except KeyError:
            if self._si._gis.users.me.username != self._si.owner:
                fl_url = self._stk_url
            else:
                fl_url = self._ssi_url

        params = {
            "outputFormat": output_format,
            "queryParameters": where,
            "portalUrl": self._si._gis._url,
            "templateItemId": report_template.id,
            "outputReportName": report_title,
            "outputPackageName": package_name,
            "surveyItemId": self._si.id,
            "featureLayerUrl": fl_url,
            "utcOffset": utc_offset,
            "uploadInfo": json.dumps(None),
            "f": "json",
            "username": self._si._gis.users.me.username,
            "locale": locale,
        }
        if merge_files:
            params["mergeFiles"] = merge_files
        if map_scale and isinstance(map_scale, (int, float)):
            params["mapScale"] = map_scale
        if webmap_item and isinstance(webmap_item, Item):
            params["webmapItemId"] = webmap_item.itemid
        if survey_item and isinstance(survey_item, Item):
            params["surveyItemId"] = survey_item.itemid
        if merge_files == "nextPage" or merge_files == "continuous":
            params["package_name"] = ""
        if folder_id:
            params["uploadInfo"] = json.dumps(
                {
                    "type": "arcgis",
                    "packageFiles": True,
                    "parameters": {"folderId": folder_id},
                }
            )
        # 1). Submit the request.
        submit = self._si._gis._con.post(
            url, params, add_headers={"X-Survey123-Request-Source": "API/Python"}
        )
        return self._check_status(
            res=submit, status_type="generate_report", save_folder=save_folder
        )

    # ----------------------------------------------------------------------
    @property
    def report_templates(self) -> list:
        """
        Returns a list of saved report items

        :returns: list of :class:`Items <arcgis.gis.Item>`
        """
        related_items = self._si.related_items(
            direction="forward", rel_type="Survey2Data"
        )
        report_templates = [t for t in related_items if t.type == "Microsoft Word"]

        return report_templates

    @property
    def reports(self) -> list:
        """returns a list of generated reports"""
        return self._si._gis.content.search(
            'owner: %s AND type:"Microsoft Word" AND tags:"Survey 123"'
            % self._ssi._gis.users.me.username,
            max_items=10000,
            outside_org=False,
        )

    # ----------------------------------------------------------------------
    def create_report_template(
        self,
        template_type: Optional[str] = "individual",
        template_name: Optional[str] = None,
        save_folder: Optional[str] = None,
    ):
        """
        The `create_report_template` creates a simple default template that
        can be downloaded locally, edited and uploaded back up as a report
        template.

        ================  ===============================================================
        **Parameter**      **Description**
        ----------------  ---------------------------------------------------------------
        template_type     Optional String. Specify which sections to include in the template.
                          Acceptable types are `individual`, `summary`, and `summaryIndividual`.
                          Default is `individual`.
        ----------------  ---------------------------------------------------------------
        template_name     Optional String. Specify the name of the output template file without file extension.
        ----------------  ---------------------------------------------------------------
        save_folder       Optional String. Specify the folder location where the output file should be stored.
        ================  ===============================================================

        :returns: String
        """
        if self._si._gis.users.me.username != self._si.owner:
            raise TypeError("Stakeholders cannot create report templates")
        try:
            if self._ssi_layers[0].properties["isView"] == True:
                fl_url = self.parent_fl_url
        except KeyError:
            fl_url = self._ssi_url

        if template_name:
            file_name = f"{template_name}.docx"
        else:
            if template_type == "individual":
                type = "Individual"
            elif template_type == "summary":
                type = "Summary"
            elif template_type == "summaryIndividual":
                type = "SummaryIndividual"
            file_name = f"{self._si.title}_sampleTemplate{type}.docx"

        url = "https://{base}/api/featureReport/createSampleTemplate".format(
            base=self._baseurl
        )
        gis = self._si._gis
        params = {
            "featureLayerUrl": fl_url,
            "surveyItemId": self._si.id,
            "portalUrl": gis._url,
            "contentType": template_type,
            "username": gis.users.me.username,
            "f": "json",
        }

        res = gis._con.post(
            url,
            params,
            try_json=False,
            out_folder=save_folder,
            file_name=file_name,
            add_headers={"X-Survey123-Request-Source": "API/Python"},
        )
        return res

    # ----------------------------------------------------------------------

    def check_template_syntax(self, template_file: Optional[str] = None):
        """
        A sync operation to check any syntax which will lead to a failure
        when generating reports in the given feature.

        ================  ===============================================================
        **Parameter**      **Description**
        ----------------  ---------------------------------------------------------------
        template_file     Required String. The report template file which syntax to be checked.
        ================  ===============================================================

        :returns: dictionary {Success or Failure}
        """

        if self._si._gis.users.me.username != self._si.owner:
            raise TypeError("Stakeholders cannot create report templates")

        try:
            if self._ssi_layers[0].properties["isView"] == True:
                fl_url = self.parent_fl_url
        except KeyError:
            fl_url = self._ssi_url

        url = "https://{base}/api/featureReport/checkTemplateSyntax".format(
            base=self._baseurl
        )
        file = {
            "templateFile": (os.path.basename(template_file), open(template_file, "rb"))
        }
        gis = self._si._gis
        params = {
            "featureLayerUrl": fl_url,
            "surveyItemId": self._si.id,
            "portalUrl": self._si._gis._url,
            "f": "json",
        }

        check = gis._con.post(
            url,
            params,
            files=file,
            add_headers={"X-Survey123-Request-Source": "API/Python"},
        )
        return check

    # ----------------------------------------------------------------------

    def upload_report_template(
        self, template_file: Optional[str] = None, template_name: Optional[str] = None
    ):
        """
        Check report template syntax to identify any syntax which will lead to a failure
        when generating reports in the given feature. Uploads the report to the organization
        and associates it with the survey.

        ================  ===============================================================
        **Parameter**      **Description**
        ----------------  ---------------------------------------------------------------
        template_file     Required String. The report template file which syntax to be checked, and uploaded.
        ----------------  ---------------------------------------------------------------
        template_name     Optional String. If provided the resulting item will use the provided name, otherwise
                          the name of the docx file will be used.
        ================  ===============================================================

        :returns: item {Success) or string (Failure}
        """

        check = self.check_template_syntax(template_file)

        if check["success"] == True:
            if template_name:
                file_name = template_name
            else:
                file_name = os.path.splitext(os.path.basename(template_file))[0]

            properties = {
                "title": file_name,
                "type": "Microsoft Word",
                "tags": "Survey123,Print Template,Feature Report Template",
                "typeKeywords": "Survey123,Survey123 Hub,Print Template,Feature Report Template",
                "snippet": "Report template",
            }
            survey_folder_id = self._si.ownerFolder
            gis = self._si._gis
            user = gis.users.get(gis.properties.user.username)
            user_folders = user.folders
            survey_folder = next(
                (f for f in user_folders if f["id"] == survey_folder_id), 0
            )
            folder = survey_folder["title"]
            # folder = "Survey-" + self._si.title
            template_item = gis.content.add(
                item_properties=properties, data=template_file, folder=folder
            )
            add_relationship = self._si.add_relationship(template_item, "Survey2Data")
        else:
            return check["details"][0]["description"]

        return template_item

    # ----------------------------------------------------------------------

    def update_report_template(self, template_file: Optional[str] = None):
        """
        Check report template syntax to identify any syntax which will lead to a failure
        when generating reports in the given feature and updates existing Report template Org item.

        ================  ===============================================================
        **Parameter**      **Description**
        ----------------  ---------------------------------------------------------------
        template_file     Required String. The report template file which syntax to be checked, and uploaded.
                          The updated template name must match the name of the existing template item.
        ================  ===============================================================

        :returns: item {Success) or string (Failure}
        """

        check = self.check_template_syntax(template_file)

        if check["success"] == True:
            file_name = os.path.splitext(os.path.basename(template_file))[0]
            gis = self._si._gis
            template_item = gis.content.search(
                query="title:" + file_name, item_type="Microsoft Word"
            )
            update = template_item[0].update(item_properties={}, data=template_file)
        else:
            return check["details"][0]["description"]

        return template_item

    # ----------------------------------------------------------------------

    def estimate(self, report_template: Item, where: str = "1=1"):
        """
        An operation to estimate how many credits are required for a task
        with the given parameters.

        ================  ===============================================================
        **Parameter**      **Description**
        ----------------  ---------------------------------------------------------------
        report_template   Required :class:`~arcgis.gis.Item` .  The report template Item.
        ----------------  ---------------------------------------------------------------
        where             Optional String. This is the select statement used to export
                          part or whole of the dataset. If the filtered result has more
                          than one feature/record, the request will be considered as a
                          batch printing. Currently, one individual report will be
                          generated for each feature/record.
        ================  ===============================================================

        :returns: dictionary {totalRecords, cost(in credits)}
        """
        try:
            if (
                self._si._gis.users.me.username == self._si.owner
                and self._ssi_layers[0].properties["isView"] == True
            ):
                fl_url = self.parent_fl_url
            elif self._si._gis.users.me.username != self._si.owner:
                fl_url = self._stk_url
        except KeyError:
            if self._si._gis.users.me.username != self._si.owner:
                fl_url = self._stk_url
            else:
                fl_url = self._ssi_url

        gis = self._si._gis
        if isinstance(where, str):
            where = {"where": where}

        url = "https://{base}/api/featureReport/estimateCredits".format(
            base=self._baseurl
        )
        params = {
            "featureLayerUrl": fl_url,
            "queryParameters": where,
            "templateItemId": report_template.id,
            "surveyItemId": self._si.id,
            "portalUrl": self._si._gis._url,
            "f": "json",
        }

        estimate = gis._con.get(
            url, params, add_headers={"X-Survey123-Request-Source": "API/Python"}
        )
        return estimate

    # ----------------------------------------------------------------------

    def create_sample_report(
        self,
        report_template: Item,
        where: str = "1=1",
        utc_offset: str = "+00:00",
        report_title: Optional[str] = None,
        merge_files: Optional[str] = None,
        survey_item: Optional[Item] = None,
        webmap_item: Optional[Item] = None,
        map_scale: Optional[float] = None,
        locale: str = "en",
        save_folder: Optional[str] = None,
    ) -> str:
        """
        Similar task to generate_report for creating test sample report, and refining
        a report template before generating any formal report.

        ================  ===============================================================
        **Parameter**      **Description**
        ----------------  ---------------------------------------------------------------
        report_template   Required :class:`~arcgis.gis.Item`. The report template Item.
        ----------------  ---------------------------------------------------------------
        where             Optional String. This is the select statement used to export
                          part or whole of the dataset.  If the record count is > 1, then
                          the item must be saved to your organization.
        ----------------  ---------------------------------------------------------------
        utc_offset        Optional String.  This is the time offset from UTC to match the
                          users timezone. Example: EST - "+04:00"
        ----------------  ---------------------------------------------------------------
        report_title      Optional String. An :class:`~arcgis.gis.Item` with this argument
                          as the title if no `save_folder` argument. If `save_folder`
                          argument is provided, this argument will be the name of the
                          output file, or the base name for files in the output zipped
                          package if the server-side component chose to zip up the output
                          (depends upon the size and number of files that would result).

                          .. note::
                              If `merge_files` is either `nextPage` or `continuous`,
                              `report_title` is the output file name.
        ----------------  ---------------------------------------------------------------
        merge_files       Optional String. Specify if output is a single file containing individual
                          records on multiple pages (`nextPage` or `continuous`) or
                          multiple files (`none`).

                          + `none` - Print multiple records in split mode. Each record
                            is a separate file. This is the default value.
                          + `nextPage` - Print multiple records in a single document.
                            Each record starts on a new page.
                          + `continuous` - Print multiple records in a single document.
                            EAch records starts on the same page of the previous record.

                          .. note::
                              A merged file larger than 500 MB will be split into multiple
                              files.
        ----------------  ---------------------------------------------------------------
        save_folder       Optional String. Specify the folder location where the output
                          file should be stored.
        ----------------  ---------------------------------------------------------------
        survey_item       Optional survey :class:`~arcgis.gis.Item` to provide additional
                          information on the survey structure.
        ----------------  ---------------------------------------------------------------
        webmap_item       Optional :class:`~arcgis.gis.Item` . Specify the base map for printing task when printing
                          a point/polyline/polygon. This takes precedence over the map set for
                          each question inside a survey.
        ----------------  ---------------------------------------------------------------
        map_scale         Optional Float. Specify the map scale when printing, the map will center on the feature geometry.
        ----------------  ---------------------------------------------------------------
        locale            Optional String. Specify the locale setting to format number and date values.
        ================  ===============================================================

        :Returns: String

        """
        try:
            if (
                self._si._gis.users.me.username == self._si.owner
                and self._ssi_layers[0].properties["isView"] == True
            ):
                fl_url = self.parent_fl_url
            elif self._si._gis.users.me.username != self._si.owner:
                fl_url = self._stk_url
        except KeyError:
            if self._si._gis.users.me.username != self._si.owner:
                fl_url = self._stk_url
            else:
                fl_url = self._ssi_url

        if isinstance(where, str):
            where = {"where": where}

        url = "https://{base}/api/featureReport/createSampleReport/submitJob".format(
            base=self._baseurl
        )

        params = {
            "queryParameters": where,
            "portalUrl": self._si._gis._url,
            "templateItemId": report_template.id,
            "surveyItemId": self._si.id,
            "featureLayerUrl": fl_url,
            "utcOffset": utc_offset,
            "f": "json",
            "locale": locale,
        }
        if merge_files:
            params["mergeFiles"] = merge_files
        if map_scale and isinstance(map_scale, (int, float)):
            params["mapScale"] = map_scale
        if webmap_item and isinstance(webmap_item, Item):
            params["webmapItemId"] = webmap_item.itemid
        if survey_item and isinstance(survey_item, Item):
            params["surveyItemId"] = survey_item.itemid
        if merge_files == "nextPage" or merge_files == "continuous":
            params["package_name"] = ""
        if report_title:
            params["outputReportName"] = report_title

        # 1). Submit the request.
        submit = self._si._gis._con.post(
            url, params, add_headers={"X-Survey123-Request-Source": "API/Python"}
        )
        return self._check_status(
            res=submit, status_type="generate_report", save_folder=save_folder
        )

    # ----------------------------------------------------------------------

    def _check_status(self, res, status_type, save_folder):
        """checks the status of a Survey123 operation"""
        jid = res["jobId"]
        gis = self._si._gis
        params = {
            "f": "json",
            "username": self._si._gis.users.me.username,
            "portalUrl": self._si._gis._url,
        }
        status_url = "https://{base}/api/featureReport/jobs/{jid}/status".format(
            base=self._baseurl, jid=jid
        )
        # 3). Start Checking the status
        res = gis._con.get(
            status_url,
            params=params,
            add_headers={"X-Survey123-Request-Source": "API/Python"},
        )
        while res["jobStatus"] == "esriJobExecuting":
            res = self._si._gis._con.get(
                status_url,
                params=params,
                add_headers={"X-Survey123-Request-Source": "API/Python"},
            )
            time.sleep(1)
        if status_type == "default_report_template":
            if (
                "results" in res
                and "details" in res["results"]
                and "resultFile" in res["results"]["details"]
                and "url" in res["results"]["details"]["resultFile"]
            ):
                url = res["results"]["details"]["resultFile"]["url"]
                file_name = os.path.basename(url)
                return gis._con.get(url, file_name=file_name, out_folder=save_folder)
            return res
        elif status_type == "generate_report":
            urls = []
            files = []
            items = []
            if res["jobStatus"] == "esriJobSucceeded":
                if "resultFiles" in res["resultInfo"]:
                    for sub in res["resultInfo"]["resultFiles"]:
                        if "id" in sub:
                            items.append(sub["id"])
                        elif "url" in sub:
                            urls.append(sub["url"])
                    files = [
                        self._si._gis._con.get(
                            url,
                            file_name=os.path.basename(urlparse(url).path),
                            add_token=False,
                            try_json=False,
                            out_folder=save_folder,
                        )
                        for url in urls
                    ] + [gis.content.get(i) for i in items]
                    if len(files) == 1:
                        return files[0]
                    return files
                elif "details" in res["resultInfo"]:
                    for res in res["resultInfo"]["details"]:
                        if "resultFile" in res:
                            fr = res["resultFile"]
                            if "id" in fr:
                                items.append(fr["id"])
                            else:
                                urls.append(fr["url"])
                        del res

                    files = [
                        self._si._gis._con.get(
                            url, file_name=os.path.basename(url), out_folder=save_folder
                        )
                        for url in urls
                    ] + [gis.content.get(i) for i in items]
                    if len(files) == 1:
                        return files[0]
                    else:
                        return files
            elif (
                res["jobStatus"] == "esriJobPartialSucceeded"
                or res["jobStatus"] == "esriJobFailed"
            ):
                raise ServerError(res["messages"][0])
            # return

    # ----------------------------------------------------------------------
    def _find_parent(self, view_url):
        """Finds the parent feature layer for a feature layer view"""
        url = view_url + "sources"
        response = self._si._gis._con.get(url)
        return response["services"][0]["url"]

    # ----------------------------------------------------------------------
    def _find_layer_name(self):
        """Finds the name of the layer the survey is submitting to, used to find the appropriate layer index"""
        name = self._si._gis._con.get(
            f"{self._gis._url}/sharing/rest/content/items/{self._si.id}/info/forminfo.json"
        )["name"]
        title = quote(name, safe="()!-_.'~")
        url = f"{self._gis._url}/sharing/rest/content/items/{self._si.id}/info/{title}.xml"
        response = self._si._gis._con.get(url, out_folder=tempfile.gettempdir())
        tree = ET.parse(response)
        root = tree.getroot()
        for elem in root[0][1].iter():
            for key, value in zip(elem.attrib.keys(), elem.attrib.values()):
                if key == "id":
                    return value



import os
import math
import numpy as np
import torch
from fastai.vision.data import ObjectCategoryList, ObjectItemList
from fastai.vision.image import ImageBBox
from fastai.core import split_kwargs_by_func, has_arg
from fastai.vision import imagenet_stats
from torchvision.transforms import Normalize
from .common import (
    ArcGISMSImage,
    get_nbatches,
    denorm_x,
    dynamic_range_adjustment,
    image_batch_stretcher,
)
from .utils import check_imbalance
from matplotlib import pyplot as plt
from matplotlib import patheffects
from fastai.basic_data import DatasetType
from fastai.torch_core import grab_idx
from .._utils.env import is_arcgispronotebook
from typing import Callable
import warnings


class ObjectDetectionCategoryList(ObjectCategoryList):
    "`ItemList` for labelled bounding boxes detected using YOLOv3."

    def analyze_pred(
        self,
        pred,
        model,
        thresh=0.5,
        nms_overlap=0.1,
        ret_scores=True,
        device=torch.device("cpu"),
    ):
        return model._analyze_pred(
            pred,
            thresh=thresh,
            nms_overlap=nms_overlap,
            ret_scores=ret_scores,
            device=device,
        )

    def reconstruct(self, t, x):
        return _reconstruct(t, x, self.pad_idx, self.classes)


class ObjectDetectionItemList(ObjectItemList):
    "`ItemList` suitable for object detection."
    _label_cls, _square_show_res = ObjectDetectionCategoryList, False
    _div = None
    _imagery_type = None

    def open(self, fn):
        return ArcGISMSImage.open(fn, div=self._div, imagery_type=self._imagery_type)

    def check_class_imbalance(
        self, func: Callable, stratify=False, class_imbalance_pct=0.01
    ):
        try:
            labelval = [(func(o)[-1]) for o in self.items]
            total_sample = np.concatenate(labelval)
            unique_sample = set(total_sample)
            check_imbalance(total_sample, unique_sample, class_imbalance_pct, stratify)
        except Exception as e:
            warnings.warn(f"Unable to check for class imbalance [reason : {e}]")

        return self

    def label_list_from_func(self, func: Callable):
        "Apply `func` to every input to get its label."
        import pandas as pd

        self._list_of_labels = ["_".join(func(o)) for o in self.items]
        self._idx_label_tuple_list = [
            (i, label) for i, label in enumerate(self._list_of_labels)
        ]
        label_series = pd.Series(self._list_of_labels)
        single_instance_labels = list(
            label_series.value_counts()[label_series.value_counts() == 1].index
        )
        self._label_idx_mapping = {
            label: i for i, label in enumerate(self._list_of_labels)
        }
        for (
            label
        ) in single_instance_labels:  # adding duplicate instance of unique labels
            self._idx_label_tuple_list.append((self._label_idx_mapping[label], label))
        return self

    def stratified_split_by_pct(self, valid_pct: float = 0.2, seed: int = None):
        try:
            "Split the items in a stratified manner by putting `valid_pct` in the validation set, optional `seed` can be passed."
            from sklearn.model_selection import train_test_split
            import random, math

            if valid_pct == 0.0:
                return self.split_none()
            if seed is not None:
                np.random.seed(seed)
            if (
                len(set(self._list_of_labels)) > len(self._list_of_labels) * valid_pct
            ):  # if validation samples length is less than unique labels
                classes = len(set(self._list_of_labels))
                xlen = len(self._list_of_labels)
                sample_shortage = math.ceil((classes - xlen * valid_pct) / valid_pct)
                extra_samples = random.choices(
                    self._idx_label_tuple_list, k=sample_shortage
                )
                self._idx_label_tuple_list.extend(extra_samples)
            X, y = [], []
            for index, label in self._idx_label_tuple_list:
                X.append(index)
                y.append(label)
            train_idx, val_idx, _, _ = train_test_split(
                X, y, test_size=valid_pct, random_state=seed, stratify=y
            )
            return self.split_by_idxs(train_idx, val_idx)
        except Exception as e:
            warnings.warn(
                f"Unable to perform stratified splitting [reason : {e}], falling back to random split"
            )
            return self.split_by_rand_pct(valid_pct=valid_pct, seed=seed)


def _reconstruct(t, x, pad_idx, classes):
    """Function to take post-processed output of model and return ImageBBox."""

    if t is None:
        return None

    t = list(t)
    if len(t[0]) == 0:
        return None

    if len(t) == 3:
        bboxes, labels, scores = t
        if len((labels - pad_idx).nonzero()) == 0:
            ret = ImageBBox.create(
                *x.size, bboxes, labels=labels, classes=classes, scale=False
            )
            ret.scores = t[2]
            return ret
        i = (labels - pad_idx).nonzero().min()
        bboxes, labels, scores = bboxes[i:], labels[i:], scores[i:]
        ret = ImageBBox.create(
            *x.size, bboxes, labels=labels, classes=classes, scale=False
        )
        ret.scores = t[2]
        return ret
    else:
        bboxes, labels = t
        if len((labels - pad_idx).nonzero()) == 0:
            return ImageBBox.create(
                *x.size, bboxes, labels=labels, classes=classes, scale=False
            )
        i = (labels - pad_idx).nonzero().min()
        bboxes, labels = bboxes[i:], labels[i:]
        return ImageBBox.create(
            *x.size, bboxes, labels=labels, classes=classes, scale=False
        )


class ObjectMSItemList(ObjectItemList):
    "`ItemList` suitable for object detection on Multispectral Data."
    _label_cls, _square_show_res = ObjectDetectionCategoryList, False

    def open(self, fn):
        return ArcGISMSImage.open_gdal(fn)


def show_batch_object_detection(
    self, rows=5, ds_type=DatasetType.Train, reverse=False, **kwargs
):
    """
    This function randomly picks a few training chips and visualizes them.
    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    rows                    Optional Integer.
                            Number of rows to display.
                            Default: 5.
    ---------------------   -------------------------------------------
    reverse                 Optional Boolean.
                            'reverse' argument flips the batch when True,
                            Default: False.
    -------------------------------------------------------------------
    """
    x, y = self.one_batch(ds_type, True, True)
    if reverse:
        x, y = x.flip(0), (y[0].flip(0), y[1].flip(0))
    n_items = rows**2 if self.train_ds.x._square_show else rows
    if self.dl(ds_type).batch_size < n_items:
        n_items = self.dl(ds_type).batch_size
    xs = [self.train_ds.x.reconstruct(grab_idx(x, i)) for i in range(n_items)]
    if has_arg(self.train_ds.y.reconstruct, "x"):
        ys = [
            self.train_ds.y.reconstruct(grab_idx(y, i), x=x) for i, x in enumerate(xs)
        ]
    else:
        ys = [self.train_ds.y.reconstruct(grab_idx(y, i)) for i in range(n_items)]
    self.train_ds.x.show_xys(xs, ys, **kwargs)


def show_batch_pascal_voc_rectangles(
    self, rows=3, alpha=1, **kwargs
):  # parameters adjusted in kwargs
    """
    This function randomly picks a few training chips and visualizes them.
    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    rows                    Optional Integer.
                            Number of rows to display.
                            Default: 3.
    ---------------------   -------------------------------------------
    alpha                   Optional Float.
                            Opacity of the lables for the corresponding
                            images. Values range between 0 and 1, where
                            1 means opaque.
    -------------------------------------------------------------------
    """
    nrows = rows
    ncols = kwargs.get("ncols", nrows)
    # start_index = kwargs.get('start_index', 0) # Does not work with dataloader

    n_items = kwargs.get("n_items", nrows * ncols)
    n_items = min(n_items, len(self.x))
    nrows = math.ceil(n_items / ncols)

    type_data_loader = kwargs.get(
        "data_loader", "training"
    )  # options : traininig, validation, testing
    if type_data_loader == "training":
        data_loader = self.train_dl
    elif type_data_loader == "validation":
        data_loader = self.valid_dl
    elif type_data_loader == "testing":
        data_loader = self.test_dl
    else:
        e = Exception(f"could not find {type_data_loader} in data.")
        raise (e)

    rgb_bands = kwargs.get("rgb_bands", self._symbology_rgb_bands)
    nodata = kwargs.get("nodata", 0)
    imsize = kwargs.get("imsize", 5)
    statistics_type = kwargs.get(
        "statistics_type", "dataset"
    )  # Accepted Values `dataset`, `DRA`
    stretch_type = kwargs.get(
        "stretch_type", "minmax"
    )  # Accepted Values `minmax`, `percentclip`
    label_font_size = kwargs.get("label_font_size", 16)

    e = Exception(
        "`rgb_bands` should be a valid band_order, list or tuple of length 3 or 1."
    )
    symbology_bands = []
    if not (len(rgb_bands) == 3 or len(rgb_bands) == 1):
        raise (e)
    for b in rgb_bands:
        if type(b) == str:
            b_index = self._bands.index(b)
        elif type(b) == int:
            self._bands[
                b
            ]  # To check if the band index specified by the user really exists.
            b_index = b
        else:
            raise (e)
        b_index = self._extract_bands.index(b_index)
        symbology_bands.append(b_index)

    # Get Batch
    x_batch, y_batch = get_nbatches(data_loader, math.ceil(n_items / self.batch_size))
    x_batch = torch.cat(x_batch)

    # Denormalize X
    x_batch = denorm_x(x_batch, self)

    y_bboxes = []
    y_classes = []
    for yb in y_batch:
        y_bboxes.extend(yb[0])
        y_classes.extend(yb[1])
    # return y_bboxes, y_classes, x_batch

    # Extract N Items and RGB Bands
    symbology_x_batch = x_batch[: (nrows * ncols), symbology_bands]
    if stretch_type is not None:
        symbology_x_batch = image_batch_stretcher(
            symbology_x_batch, stretch_type, statistics_type
        )
        # symbology_x_batch = dynamic_range_adjustment(symbology_x_batch)

    # Channel first to channel last and clamp float values to range 0 - 1 for plotting
    symbology_x_batch = symbology_x_batch.permute(0, 2, 3, 1)
    # Clamp float values to range 0 - 1
    if symbology_x_batch.mean() < 1:
        symbology_x_batch = symbology_x_batch.clamp(0, 1)

    # Squeeze channels if single channel (1, 224, 224) -> (224, 224)
    if symbology_x_batch.shape[-1] == 1:
        symbology_x_batch = symbology_x_batch.squeeze()

    # Get color Array
    color_array = self._multispectral_color_array
    color_array[1:, 3] = alpha

    # Size for plotting
    fig, axs = plt.subplots(
        nrows=nrows, ncols=ncols, figsize=(ncols * imsize, nrows * imsize)
    )
    idx = 0
    for r in range(nrows):
        for c in range(ncols):
            axi = axs
            if nrows == 1:
                axi = axi
            else:
                axi = axi[r]
            if ncols == 1:
                axi = axi
            else:
                axi = axi[c]
            if idx < symbology_x_batch.shape[0]:
                axi.imshow(symbology_x_batch[idx].cpu().numpy())
                classes = y_classes[idx][y_classes[idx] > 0]
                bboxes = y_bboxes[idx][y_classes[idx] > 0]
                bboxes = (bboxes + 1) * 0.5
                bboxes = bboxes.clamp(0, 1) * (x_batch.shape[-1] - 1)
                for i, bbox in enumerate(bboxes):
                    xs = bbox[[1, 1, 3, 3, 1]]
                    ys = bbox[[0, 2, 2, 0, 0]]
                    color = self._multispectral_color_array[classes[i]].tolist()
                    axi.plot(
                        xs.cpu().numpy(), ys.cpu().numpy(), color=color, linewidth=2
                    )
                    class_value = classes[i].item()
                    lbl = self.classes[class_value]
                    if lbl.strip() == "":
                        lbl = str(class_value)
                    axi.text(
                        xs[0] + 1,
                        ys[0] + 1 + (label_font_size * (x_batch.shape[-1] - 1) / 256),
                        lbl,
                        size=label_font_size,
                        color=color,
                        path_effects=[
                            patheffects.Stroke(linewidth=0.5, foreground="gray")
                        ],
                    )
                axi.axis("off")
            else:
                axi.axis("off")
            idx += 1
    if is_arcgispronotebook():
        plt.show()


def show_results_multispectral(
    self, nrows=5, alpha=1, **kwargs
):  # parameters adjusted in kwargs
    from matplotlib import pyplot as plt
    from matplotlib import patheffects

    # Get Number of items
    ncols = 2

    type_data_loader = kwargs.get(
        "data_loader", "validation"
    )  # options : traininig, validation, testing

    if type_data_loader == "training":
        data_loader = self._data.train_dl
    elif type_data_loader == "validation":
        data_loader = self._data.valid_dl
    elif type_data_loader == "testing":
        data_loader = self._data.test_dl
    else:
        e = Exception(
            f"could not find {type_data_loader} in data. Please ensure that the data loader type is traininig, validation or testing "
        )
        raise (e)

    nodata = kwargs.get("nodata", 0)
    return_fig = kwargs.get("return_fig", False)

    index = kwargs.get("start_index", 0)

    imsize = kwargs.get("imsize", 4)

    thresh = kwargs.get("thresh", 0.3)

    nms_overlap = kwargs.get("nms_overlap", 0.1)

    if getattr(self, "_is_model_extension", False):
        transform_kwargs, kwargs = split_kwargs_by_func(
            kwargs, self._model_conf.transform_input_multispectral
        )

    title_font_size = 16
    _top = 1 - (math.sqrt(title_font_size) / math.sqrt(100 * nrows * imsize))
    top = kwargs.get("top", _top)

    statistics_type = kwargs.get(
        "statistics_type", "dataset"
    )  # Accepted Values `dataset`, `DRA`
    stretch_type = kwargs.get(
        "stretch_type", "minmax"
    )  # Accepted Values `minmax`, `percentclip`
    label_font_size = kwargs.get("label_font_size", 16)

    # Get Batch
    x_batch, y_batch = get_nbatches(
        data_loader, math.ceil(nrows / self._data.batch_size)
    )

    if self._data.norm is None and not self._data._is_multispectral:
        normalize = Normalize(mean=imagenet_stats[0], std=imagenet_stats[1])
        modified_x_batch = []
        for i in x_batch:
            modified_x_batch.append(normalize(i))
        x_batch = modified_x_batch
        del modified_x_batch

    x_batch = torch.cat(x_batch)
    y_bboxes = []
    y_classes = []
    for yb in y_batch:
        y_bboxes.extend(yb[0])
        y_classes.extend(yb[1])

    nrows = min(nrows, len(x_batch))

    predictions_store = []
    pred_model_external = []

    for i in range(0, x_batch.shape[0], self._data.batch_size):
        if self._backend == "pytorch":
            if getattr(self, "_is_model_extension", False):
                xb = self._model_conf.transform_input_multispectral(
                    x_batch[i : i + self._data.batch_size], **transform_kwargs
                )
                try:
                    _pred_ext = self.learn.model.eval()(xb)
                except Exception as e:
                    if getattr(self, "_is_fasterrcnn", False):
                        _pred_ext = []
                        for _ in range(self._data.batch_size):
                            res = {}
                            res["boxes"] = torch.empty(0, 4)
                            res["scores"] = torch.tensor([])
                            res["labels"] = torch.tensor([])
                            _pred_ext.append(res)
                    else:
                        raise e
                analyzed_pred_ext = self._analyze_pred(
                    _pred_ext,
                    thresh=thresh,
                    nms_overlap=nms_overlap,
                    ret_scores=True,
                    device=self._device,
                )
            else:
                predictions = self.learn.model.eval()(
                    x_batch[i : i + self._data.batch_size]
                )
        elif self._backend == "tensorflow":
            from .fastai_tf_fit import _pytorch_to_tf_batch

            _classes_sparse, _activations = self.learn.model(
                _pytorch_to_tf_batch(x_batch[i : i + self._data.batch_size])
            )
            _classes_sparse, _activations = (
                _classes_sparse.detach().numpy(),
                _activations.detach().numpy(),
            )
            predictions = (torch.tensor(_classes_sparse), torch.tensor(_activations))

        if getattr(self, "_is_model_extension", False):
            pred_model_external.extend(analyzed_pred_ext)
        else:
            predictions_store.append(predictions)

    if not getattr(self, "_is_model_extension", False):
        if self.__class__.__name__ == "YOLOv3":
            predictions_store = torch.cat(predictions_store)
        else:
            __predictions_store = []
            for __batch in predictions_store:
                for __chip in zip(*__batch):
                    __predictions_store.append((__chip[0], __chip[1]))
            predictions_store = __predictions_store

    if self._is_multispectral:
        rgb_bands = kwargs.get("rgb_bands", self._data._symbology_rgb_bands)

        e = Exception(
            "`rgb_bands` should be a valid band_order, list or tuple of length 3 or 1."
        )
        symbology_bands = []
        if not (len(rgb_bands) == 3 or len(rgb_bands) == 1):
            raise (e)
        for b in rgb_bands:
            if type(b) == str:
                b_index = self._bands.index(b)
            elif type(b) == int:
                self._bands[
                    b
                ]  # To check if the band index specified by the user really exists.
                b_index = b
            else:
                raise (e)
            b_index = self._data._extract_bands.index(b_index)
            symbology_bands.append(b_index)

        # Denormalize X
        x_batch = (
            self._data._scaled_std_values[self._data._extract_bands]
            .view(1, -1, 1, 1)
            .to(x_batch)
            * x_batch
        ) + self._data._scaled_mean_values[self._data._extract_bands].view(
            1, -1, 1, 1
        ).to(
            x_batch
        )

        # Extract RGB Bands
        symbology_x_batch = x_batch[: (nrows * ncols), symbology_bands]
        if stretch_type is not None:
            symbology_x_batch = image_batch_stretcher(
                symbology_x_batch, stretch_type, statistics_type
            )
    else:
        # normalization stats
        symbology_x_batch = denorm_x(x_batch)

    # Channel first to channel last for plotting
    symbology_x_batch = symbology_x_batch.permute(0, 2, 3, 1)
    # Clamp float values to range 0 - 1
    if symbology_x_batch.mean() < 1:
        symbology_x_batch = symbology_x_batch.clamp(0, 1)

    # Squeeze channels if single channel (1, 224, 224) -> (224, 224)
    if symbology_x_batch.shape[-1] == 1:
        symbology_x_batch = symbology_x_batch.squeeze()

    # Get color Array
    color_array = self._data._multispectral_color_array
    color_array[:, 3] = alpha

    # Size for plotting
    fig, axs = plt.subplots(
        nrows=nrows, ncols=ncols, figsize=(ncols * imsize, nrows * imsize)
    )
    fig.suptitle("Ground Truth / Predictions", fontsize=title_font_size)
    plt.subplots_adjust(top=top)
    idx = 0
    for r in range(0, nrows):
        if nrows == 1:
            ax_i = axs
        else:
            ax_i = axs[r]

        # Plot Ground Truth
        ax_ground_truth = ax_i[0]
        ax_ground_truth.axis("off")
        ax_ground_truth.imshow(symbology_x_batch[idx].cpu().numpy())
        gt_classes = y_classes[idx][y_classes[idx] > 0]
        gt_bboxes = y_bboxes[idx][y_classes[idx] > 0]
        gt_bboxes = (gt_bboxes + 1) * 0.5
        gt_bboxes = gt_bboxes.clamp(0, 1) * (x_batch.shape[-1] - 1)
        for i, bbox in enumerate(gt_bboxes):
            xs = bbox[[1, 1, 3, 3, 1]]
            ys = bbox[[0, 2, 2, 0, 0]]
            color = self._data._multispectral_color_array[gt_classes[i]].tolist()
            ax_ground_truth.plot(
                xs.cpu().numpy(),
                ys.cpu().numpy(),
                color=color,
                linewidth=2,
                path_effects=[
                    patheffects.Stroke(linewidth=3, foreground="black"),
                    patheffects.Normal(),
                ],
            )
            class_value = gt_classes[i].item()
            lbl = self._data.classes[class_value]
            if lbl.strip() == "":
                lbl = str(class_value)
            ax_ground_truth.text(
                xs[0] + 1,
                ys[0] + 1 + (label_font_size * (x_batch.shape[-1] - 1) / 256),
                lbl,
                size=label_font_size,
                color=color,
                path_effects=[
                    patheffects.Stroke(linewidth=1, foreground="black"),
                    patheffects.Normal(),
                ],
            )

        # Plot Predictions
        ax_prediction = ax_i[1]
        ax_prediction.axis("off")
        ax_prediction.imshow(symbology_x_batch[idx].cpu().numpy())

        if getattr(self, "_is_model_extension", False):
            analyzed_prediction = pred_model_external[idx]
        else:
            analyzed_prediction = self._analyze_pred(
                predictions_store[idx],
                thresh=thresh,
                nms_overlap=nms_overlap,
                ret_scores=True,
                device=self._device,
            )

        if analyzed_prediction is not None:
            (
                predicted_bboxes,
                predicted_classes,
                predicted_confidences,
            ) = analyzed_prediction
            predicted_bboxes = (predicted_bboxes + 1) * 0.5
            predicted_bboxes = predicted_bboxes.clamp(0, 1) * (x_batch.shape[-1] - 1)
            if len(predicted_bboxes) > 0:
                for i, bbox in enumerate(predicted_bboxes):
                    xs = bbox[[1, 1, 3, 3, 1]]
                    ys = bbox[[0, 2, 2, 0, 0]]
                    color = self._data._multispectral_color_array[
                        predicted_classes[i]
                    ].tolist()
                    ax_prediction.plot(
                        xs.detach().cpu().numpy(),
                        ys.detach().cpu().numpy(),
                        color=color,
                        linewidth=2,
                        path_effects=[
                            patheffects.Stroke(linewidth=3, foreground="black"),
                            patheffects.Normal(),
                        ],
                    )
                    class_value = predicted_classes[i].item()
                    lblp = self._data.classes[class_value]
                    if lblp.strip() == "":
                        lblp = str(class_value)
                    ax_prediction.text(
                        xs[0] + 1,
                        ys[0] + 1 + (label_font_size * (x_batch.shape[-1] - 1) / 256),
                        lblp,
                        size=label_font_size,
                        color=color,
                        path_effects=[
                            patheffects.Stroke(linewidth=1, foreground="black"),
                            patheffects.Normal(),
                        ],
                    )

        idx += 1
    if is_arcgispronotebook():
        plt.show()
    return fig, axs


from fastai.torch_core import try_int
from numbers import Integral
from typing import Union


def modified_getitem(self, idxs: Union[int, np.ndarray]) -> "LabelList":
    "return a single (x, y) if `idxs` is an integer or a new `LabelList` object if `idxs` is a range."
    idxs = try_int(idxs)
    if isinstance(idxs, Integral):
        if self.item is None:
            x, y = self.x[idxs], self.y[idxs]
        else:
            x, y = self.item, 0
        if self.tfms or self.tfmargs:
            x = x.apply_tfms(self.tfms, **{})
        if hasattr(self, "tfms_y") and self.tfm_y and self.item is None:
            y = y.apply_tfms(self.tfms_y, **{**self.tfmargs_y, "do_resolve": False})
        if y is None:
            y = 0
        return x, y
    else:
        return self.new(self.x[idxs], self.y[idxs])



from __future__ import annotations
from typing import Optional, Union
import uuid
from enum import Enum
from arcgis.auth.tools import LazyLoader
import copy
from ._ref import templates
from arcgis.gis import GIS
import re
from dataclasses import dataclass
import tempfile

try:
    import ujson as json
except ImportError:
    import json

arcgis = LazyLoader("arcgis")
# json = LazyLoader("json")
time = LazyLoader("time")


class Templates(Enum):
    BLANKFULLSCREEN = "blank fullscreen"
    BLANKSCROLLING = "blank scrolling"
    FOLDABLE = "foldable"
    LAUNCHPAD = "launchpad"
    JEWELERYBOX = "jewelrybox"
    BILLBOARD = "billboard"
    JOURNEY = "journey"
    RIBBON = "ribbon"
    GENERAL = "general"
    INTRODUCTION = "introduction"
    GALLERY = "gallery"
    EPIC = "epic"
    SNAPSHOT = "snapshot"
    SUMMARY = "summary"
    TIMELINE = "timeline"
    SCENIC = "scenic"
    EXHIBITION = "exhibition"
    DART = "dart"
    POCKET = "pocket"
    QUICKNAVIGATION = "quick navigation"
    PARALLAX = "parallax"
    DASH = "dash"
    INDICATOR = "indicator"
    MONITOR = "monitor"
    REVEAL = "reveal"

    def preview(self, width: Optional[int] = 800, height: Optional[int] = 500):
        import threading
        import time

        def thread_delete(item):
            time.sleep(3)
            item.delete()

        try:
            temp = WebExperience(template=self.value)
            temp._item.share(everyone=True)
            temp.save(publish=True)
            from IPython.display import IFrame

            frame = IFrame(
                src=temp._item.url,
                width=width,
                height=height,
            )
            delete = threading.Thread(target=thread_delete, args=([temp._item]))
            delete.start()
            return frame

        except:
            return False


class WebExperience(object):

    """
    A Web Experience is web-based application that provides viewers with an interactive
    interface to maps, data, feature layers, and other components of the creator's design.
    Though these experiences are normally constructed via a GUI found on ArcGIS Online or
    Enterprise, this class provides users a host of supplemental options to manage experiences,
    in addition to basic creation of experiences.

    ===============     ====================================================================
    **Argument**        **Description**
    ---------------     --------------------------------------------------------------------
    item                Optional String or Item. The string for an item id or an item of type
                        'Web Experience'. If no item is passed, a new experience is created
                        and saved to your active portal.
    ---------------     --------------------------------------------------------------------
    gis                 Optional instance of :class:`~arcgis.gis.GIS`. If none provided the active gis is used.
    ---------------     --------------------------------------------------------------------
    template            Optional string. If a new experience is being created, the template
                        used to construct the layout. If necessary and none provided, template
                        will default to `blank fullscreen`.
    ---------------     --------------------------------------------------------------------
    path                Optional string. Used if a WebExperience is being based on a local
                        config json file, as one would with Experiences made via the desktop
                        version of Experience Builder.
    ---------------     --------------------------------------------------------------------
    name                Optional string. If a new experience is being created, the name of the
                        item. Otherwise, will default to "Experience via Python" followed by a
                        random number.
    ===============     ====================================================================
    """

    _draft = {}
    _itemid = None
    _item = None
    _local = False
    _properties = None
    _gis = None
    _resources = None
    _expdict = {}

    def __init__(
        self,
        item: Optional[Union[arcgis.gis.Item, str]] = None,
        path: Optional[str] = None,
        gis: Optional[arcgis.gis.GIS] = None,
        template: Optional[Union[Templates, str]] = None,
        name: Optional[str] = None,
    ):
        if gis is None:
            if item and isinstance(item, arcgis.gis.Item):
                self._gis = item._gis
            else:
                self._gis = arcgis.env.active_gis
        else:
            if item and isinstance(item, arcgis.gis.Item):
                if item._gis != gis:
                    raise ValueError("Provided GIS must match item GIS")
            self._gis = gis

        if self._gis is None or self._gis._portal.is_logged_in is False:
            # check to see if user is authenticated
            raise AttributeError(
                "Must be logged into a Portal Account or provide a logged-in GIS object"
            )
        if item and isinstance(item, str):
            # get item using the item id
            item = self._gis.content.get(item)
            if item is None:
                raise ValueError("Item is not accessible with provided GIS")
        if item and isinstance(item, arcgis.gis.Item) and item.type == "Web Experience":
            # set item properties
            self._item = item
            self._itemid = self._item.itemid
            self._resources = self._item.resources.list()
            self._expdict = self._item.resources.get("config/config.json")
            self._draft = self._item.resources.get("config/config.json")
            self._gis = self._item._gis
        elif path and isinstance(path, str):
            # construct web experience from path's config file
            # note that this will not have an associated portal item/itemid/resources
            self._local = True
            with open(path) as json_file:
                config = json.load(json_file)
                self._draft = config
                self._expdict = config
        elif (
            item and isinstance(item, arcgis.gis.Item) and item.type != "Web Experience"
        ):
            # Throw error if item is not of type Experience
            raise ValueError("Item is not a Web Experience or is inaccesible")
        else:
            self._create_new_experience(template=template, name=name)

    # -----------------------------------------------------------------------------------
    @property
    def item(self):
        """
        Returns the portal item associated with the WebExperience, if possible.
        Experiences made from local json files won't have an `item` property until they
        are added to a portal using `upload()`.
        """
        return self._item

    # -----------------------------------------------------------------------------------
    @property
    def itemid(self):
        """
        Returns the item ID of the associated portal item, if possible. As with `item`,
        Experiences made from local json files won't have an `itemid` property until they
        are added to a portal using `upload()`.
        """
        return self._itemid

    # -----------------------------------------------------------------------------------
    @property
    def datasources(self):
        """
        Shows the data sources dictionary found in the experience's draft, allowing users
        to quickly get info on all of the other items in their experience. Changing this
        dictionary will change the draft, making it convenient for remapping data
        sources.
        """
        return self._draft["dataSources"]

    # -----------------------------------------------------------------------------------
    def _create_new_experience(
        self,
        config=None,
        template="blank fullscreen",
        name=None,
        gis=None,
        item_properties={},
    ):
        """
        If no experience is specified when creating a WebExperience, this helper function
        creates a new experience and saves it as an item to the active GIS. Users can specify
        a template from the experience builder to create their template, in addition to a custom
        item name (done as arguments in the initial creation of the WebExperience).
        """

        if config is None:
            if isinstance(template, Templates):
                template = template.value

            # retrieve template for experience
            if template is None:
                template = "blank fullscreen"

            temp_low = template.lower()
            if temp_low in arcgis.apps.expbuilder._ref.templates:
                temp_dict = copy.deepcopy(
                    arcgis.apps.expbuilder._ref.templates[temp_low]
                )
            else:
                temp_dict = copy.deepcopy(
                    arcgis.apps.expbuilder._ref.templates["blank fullscreen"]
                )

            temp_dict["attributes"]["portalUrl"] = self._gis.url
            # temp_dict["timestamp"]
            # create item and generate basic properties
            if name is None:
                title = "Experience via Python %s" % uuid.uuid4().hex[:10]
            else:
                title = name

        else:
            temp_dict = config
            temp_dict["attributes"]["portalUrl"] = self._gis.url
            if name is None:
                title = "Experience via Python %s" % uuid.uuid4().hex[:10]
            else:
                title = name

        if gis:
            temp_dict["attributes"]["portalUrl"] = gis.url
        else:
            temp_dict["attributes"]["portalUrl"] = self._gis.url

        keywords = ",".join(
            [
                "EXB Experience",
                "JavaScript",
                "Ready To Use",
                "status: Draft",
                "Web Application",
                "Web Experience",
                "Web Page",
                "Web Site",
                "expbuilderapp:python-api-" + arcgis.__version__,
            ]
        )
        props = item_properties
        props["title"] = title
        props["type"] = "Web Experience"
        props["typeKeywords"] = keywords

        """item_properties = {
            "type": "Web Experience",
            "title": title,
            "typeKeywords": keywords,
        }"""

        # add to active gis and set properties
        if gis is None:
            item = self._gis.content.add(item_properties=props)
        else:
            item = gis.content.add(item_properties=props)

        # assign to experience properties
        self._item = item
        self._itemid = item.itemid
        self._item.resources.add(
            folder_name="config", file_name="config.json", text=temp_dict
        )
        self._resources = self._item.resources.list()
        self._expdict = temp_dict
        self._draft = temp_dict

    # ----------------------------------------------------------------------
    def save(
        self,
        title: Optional[str] = None,
        tags: Optional[list] = None,
        access: str = None,
        publish: bool = False,
        duplicate: bool = False,
        include_private: Optional[bool] = None,
        item_properties: Optional[dict] = {},
    ):
        """
        This method will save your Web Experience to your active GIS. The experience will be saved
        with unpublished changes unless the `publish` parameter is set to True. Note that this is
        different from the `publish()` method in that this will save and publish the unsaved draft
        of the WebExperience object, as opposed to the already existing save state.

        The title only needs to be specified if a change is wanted, otherwise the existing title
        is used.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        title               Optional string. The new title of the WebExperience, if desired.
        ---------------     --------------------------------------------------------------------
        tags                Optional string. Updated tags for the WebExperience, if desired.
        ---------------     --------------------------------------------------------------------
        access              Optional string. The sharing setting of the WebExperience. If none
                            is specified, the current access is kept. This is used when the
                            `publish` parameter is set to True.

                            Values: `private` | `org` | `public`
        ---------------     --------------------------------------------------------------------
        publish             Optional boolean. If True, the experience is saved and also
                            published. Default is False, meaning the experience is saved with
                            unpublished changes.
        ---------------     --------------------------------------------------------------------
        duplicate           Optional boolean. If True, the experience is duplicated and a new
                            WebExperience object is saved with any specified changes to
                            title, tags, access, or publish included. Essentially functions as a
                            "Save As" method. Default is False, meaning changes are saved to the
                            original object.
        ---------------     --------------------------------------------------------------------
        include_private     Optional boolean. Only to be included when duplicate is `True`.
                            If True, the private resources of the original item will be included
                            in the new item.
        ---------------     --------------------------------------------------------------------
        item_properties     Optional dictionary. Contains a variety of properties that can be
                            set when creating a new item, much like `item.update()`. See below
                            for a table containing possible properties.
        ===============     ====================================================================


        *Key:Value Dictionary Options for Argument item_properties*

        =================  =====================================================================
        **Key**            **Value**
        -----------------  ---------------------------------------------------------------------
        description        Optional string. Description of the item.
        -----------------  ---------------------------------------------------------------------
        url                Optional string. URL to item that are based on URLs.
        -----------------  ---------------------------------------------------------------------
        snippet            Optional string. Provide a short summary (limit to max 250 characters) of the what the item is.
        -----------------  ---------------------------------------------------------------------
        accessInformation  Optional string. Information on the source of the content.
        -----------------  ---------------------------------------------------------------------
        licenseInfo        Optional string.  Any license information or restrictions regarding the content.
        -----------------  ---------------------------------------------------------------------
        culture            Optional string. Locale, country and language information.
        =================  =====================================================================


        :return: A boolean indicating the success of the operation.
        """
        if duplicate:
            new_exp = self._duplicate(
                title=title, tags=tags, include_private=include_private
            )
            new_exp.save(access=access, publish=publish)
            return new_exp

        keywords = self._item.typeKeywords
        for i in range(len(keywords)):
            if keywords[i] == "status: Published":
                keywords[i] = "status: Changed"
        props = item_properties
        if title:
            props["title"] = title
        if tags:
            props["tags"] = tags

        self._expdict = self._draft
        # Create a temporary file and write data to it
        with tempfile.NamedTemporaryFile(
            mode="w+", suffix=".json", delete=False
        ) as tfile:
            json.dump(self._expdict, tfile)
            # Close the file explicitly
            tfile.close()
        self._item.resources.update(
            folder_name="config",
            file_name="config.json",
            file=tfile.name,
        )
        self._resources = self._item.resources.list()
        if publish:
            for i in range(len(keywords)):
                if "status" in keywords[i]:
                    keywords[i] = "status: Published"
            if access:
                props["access"] = access
            props["typeKeywords"] = keywords
            if self._gis._is_agol:
                url = "https://experience.arcgis.com/experience/" + self._item.itemid
            else:
                url = (
                    self._gis.url
                    + "/apps/experiencebuilder/experience/?id="
                    + self._item.itemid
                )
            props["url"] = url
            return self._item.update(item_properties=props, data=self._expdict)
            # self._publish(item_properties = item_properties, data = self._expdict)
        else:
            props["typeKeywords"] = keywords
            return self._item.update(item_properties=props)

    # ----------------------------------------------------------------------
    def reload(self):
        """
        Resets any changes that the user has made to the last saved state. Note that
        this only applies to changes made through a Python API object, and not the GUI.

        :return: A boolean indicating the success of the operation.
        """

        self._draft = self._expdict
        # Create a temporary file and write data to it
        with tempfile.NamedTemporaryFile(
            mode="w+", suffix=".json", delete=False
        ) as tfile:
            json.dump(self._expdict, tfile)
            # Close the file explicitly
            tfile.close()
        return self._item.resources.update(
            folder_name="config", file_name="config.json", file=tfile.name
        )

    # ----------------------------------------------------------------------
    def view(self, width: Optional[int] = 800, height: Optional[int] = 500):
        """
        Shows the currently published experience, if possible. Default width is 800 and default
        height is 500. Note that this displays the actively published version of the
        WebExperience object; to visualize unsaved changes, `preview()` should be used.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        width               Optional integer. The desired width to show the preview.
        ---------------     --------------------------------------------------------------------
        height              Optional integer. The desired height to show the preview.
        ===============     ====================================================================

        .. note::

            In some cases, a dialogue box may pop up asking for credentials when calling this
            method. If the preview isn't rendering, check if pop-ups are disabled in your browser.


        :return:
            An IFrame display of the Experience if possible, else the item url is returned to be
            clicked on. If the item is unpublished, the function returns False.
        """
        keywords = self._item.typeKeywords
        if "status: Published" in keywords:
            from IPython.display import IFrame

            try:
                frame = IFrame(
                    src=self._item.url,
                    width=width,
                    height=height,
                )
                return frame
            except:
                return self._item.url
        else:
            return False

    # ----------------------------------------------------------------------
    def delete(self):
        """
        Deletes the experience and its associated item from the portal.

        :return:
            A boolean indicating the success of the operation.
        """
        item = self._gis.content.get(self._itemid)
        return item.delete()

    # ----------------------------------------------------------------------
    def _add_resource(self, file=None, resource_name=None, text=None, access="inherit"):
        """
        See :class:`~arcgis.gis.ResourceManager`
        """
        resource_manager = arcgis.gis.ResourceManager(self._item, self._gis)
        is_present = False
        if file:
            for resource in self._resources:
                if resource["resource"] in file:
                    is_present = True
                    resp = True
        properties = {
            "editInfo": {
                "editor": self._gis._username,
                "modified": str(int(time.time() * 1000)),
                "id": uuid.uuid4().hex[0:21],
                "app": "python-api",
            }
        }

        # access is inherited from item upon add, except for json where always private
        if is_present is False:
            resp = resource_manager.add(
                file=file,
                file_name=resource_name,
                text=text,
                access=access,
                properties=properties,
            )

        self._resources = self._item.resources.list()
        return resp

    # ----------------------------------------------------------------------
    def _duplicate(
        self,
        title: Optional[str] = None,
        tags: Optional[Union[list[str], str]] = None,
        include_private: Optional[bool] = None,
    ):
        """
        Creates a copy of the experience within the active GIS. Returns a new
        WebExperience object that retains the unsaved changes from the original.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        title               Optional string. The name of the new experience. If left blank, the
                            new item copies the name of the original.
        ---------------     --------------------------------------------------------------------
        tags                Optional string. The desired tags for the new item, separated by
                            commas.
        ---------------     --------------------------------------------------------------------
        include_private     Optional boolean. If True, the private resources of the original
                            item will be included in the new item.
        ===============     ====================================================================

        :return:
            The newly created WebExperience object.

        """
        new_item = self._item.copy_item(
            title=title,
            tags=tags,
            include_resources=True,
            include_private=include_private,
        )
        if new_item:
            new_exp = WebExperience(new_item)
            new_exp._draft = self._draft
            return new_exp
        else:
            return False

    # ----------------------------------------------------------------------
    def upload(
        self,
        gis: Optional[GIS] = None,
        publish=False,
        title=None,
        item_mapping: Optional[dict] = None,
        auto_remap: Optional[bool] = False,
        item_properties: Optional[dict] = {},
    ):
        """
        Adds a WebExperience created locally through the Developer Edition to a specified
        portal. After doing this, the WebExperience object will obtain item and itemid
        properties. Gives users options to remap their experience's datasources to items
        present in the portal, both manually or automatically

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        gis                 Optional GIS object. The portal to add the WebExperience to. If none
                            is passed in, will default to the GIS of the WebExperience object.
        ---------------     --------------------------------------------------------------------
        publish             Optional boolean. Publishes the experience when adding it to the
                            portal. Default is `False`.
        ---------------     --------------------------------------------------------------------
        title               Optional string. Allows a user to specify the title of their new
                            experience in the portal.
        ---------------     --------------------------------------------------------------------
        item_mapping        Optional dictionary. Allows users to manually remap the datasources
                            of their experience to datasources present in the portal. See
                            example dictionary below.
        ---------------     --------------------------------------------------------------------
        auto_remap          Optional boolean. Searches the portal for matching datasources and
                            automatically remaps the experience to use those accordingly.
                            Default is `False`.
        ---------------     --------------------------------------------------------------------
        item_properties     Optional dictionary. Contains a variety of properties that can be
                            set when creating a new item, much like `ContentManager.add()`. See
                            below for a table containing possible properties.
        ===============     ====================================================================


        *Key:Value Dictionary Options for Argument item_properties*

        ==========================  =====================================================================
        **Key**                     **Value**
        --------------------------  ---------------------------------------------------------------------
        description                 Optional string. Description of the item.
        --------------------------  ---------------------------------------------------------------------
        url                         Optional string. URL to item that are based on URLs.
        --------------------------  ---------------------------------------------------------------------
        tags                        Optional string. Tags listed as comma-separated values, or a list of strings.
                                    Used for searches on items.
        --------------------------  ---------------------------------------------------------------------
        snippet                     Optional string. Provide a short summary (limit to max 250 characters) of the what the item is.
        --------------------------  ---------------------------------------------------------------------
        accessInformation           Optional string. Information on the source of the content.
        --------------------------  ---------------------------------------------------------------------
        licenseInfo                 Optional string.  Any license information or restrictions regarding the content.
        --------------------------  ---------------------------------------------------------------------
        culture                     Optional string. Locale, country and language information.
        --------------------------  ---------------------------------------------------------------------
        commentsEnabled             Optional boolean. Default is true, controls whether comments are allowed (true)
                                    or not allowed (false).
        --------------------------  ---------------------------------------------------------------------
        access                      Optional string. Valid values are private, org, or public. Defaults to private.
        --------------------------  ---------------------------------------------------------------------
        overwrite                   Optional boolean. Default is `false`. Controls whether item can be overwritten.
        ==========================  =====================================================================

        :return:
            The newly added portal item, if successful. Otherwise, returns False.
        """
        if gis is None:
            gis = self._gis
        # this will be the config file that dictates our new portal experience
        new_config = self._draft
        # automatically remap data to matching sources in target GIS
        if auto_remap:
            sources = new_config["dataSources"]
            for source in sources:
                # first, see if we can already access each one anonymously
                # or through the passed in GIS
                url = sources[source]["portalUrl"]
                test_gis = GIS(url=url)
                try:
                    try:
                        targ_item = test_gis.content.get(sources[source]["itemId"])
                    except:
                        targ_item = gis.content.get(sources[source]["itemId"])
                    assert targ_item
                # this will only get triggered if we can't access and need to remap
                except:
                    # if available, match item in target GIS based on title and type
                    source_title = sources[source]["sourceLabel"]
                    source_type = sources[source]["type"]
                    source_reg = re.sub("[^A-Za-z0-9]+", "", source_type).lower()
                    query = 'title:"' + source_title + '"'

                    for item in gis.content.search(query=query):
                        target_title = item.title
                        target_type = item.type
                        target_reg = re.sub("[^A-Za-z0-9]+", "", target_type).lower()

                        # must be exact match on both
                        if source_title == target_title and source_reg == target_reg:
                            new_config["dataSources"][source]["itemId"] = item.id
                            new_config["dataSources"][source]["portalUrl"] = gis.url
                            break

        # if user passed in their own custom remapping, use that
        if item_mapping is not None:
            for source, v in item_mapping.items():
                for attr, new_value in v.items():
                    new_config["dataSources"][source][attr] = new_value

        # create a new portal experience using the config
        self._create_new_experience(
            config=new_config, name=title, gis=gis, item_properties=item_properties
        )
        self._local = False

        # if wish for item to be published, save/publish
        if publish:
            self.save(publish=True)

        if self._item:
            return self._item
        else:
            return False

    # ----------------------------------------------------------------------
    def preview(self, width: Optional[int] = 800, height: Optional[int] = 500):
        """
        Show a preview of the current experience draft. The default is a width of 800 and height of 500.
        Note that this should be used to visualize unsaved changes to the WebExperience object; to see
        the actively published version of the object, `view()` should be used.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        width               Optional integer. The desired width to show the preview.
        ---------------     --------------------------------------------------------------------
        height              Optional integer. The desired height to show the preview.
        ===============     ====================================================================

        .. note::

            In some cases, a dialogue box may pop up asking for credentials when calling this
            method. If the preview isn't rendering, check if pop-ups are disabled in your browser.


        :return:
            An IFrame display of the Experience if possible, else the item url is returned to be
            clicked on.
        """
        import threading
        import time

        def thread_delete(item):
            time.sleep(3)
            item.delete()

        try:
            dummy_exp = self._duplicate()
            dummy_exp._item.share(everyone=True)
            dummy_exp.save(publish=True)
            from IPython.display import IFrame

            frame = IFrame(
                src=dummy_exp._item.url,
                width=width,
                height=height,
            )
            delete = threading.Thread(target=thread_delete, args=([dummy_exp._item]))
            delete.start()
            return frame

        except:
            return self._item.url

    # ----------------------------------------------------------------------
    def clone(self, target, owner, **kwargs):
        """
        Clones the experience and all of it's data sources to a target GIS. User must
        have admin privileges on the original item's GIS, and provide an authenticated
        instance of a target GIS. Users must also specify the name of an account on the
        target GIS to own the items. Also accepts arguments for :class:`~arcgis.gis.Item.clone_items()`

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        target              Required GIS. An authenticated instance of the GIS that the user
                            wishes to clone the experience to.
        ---------------     --------------------------------------------------------------------
        owner               Required string. The username of the account that will be the owner
                            of the experience and its data source items in the target GIS.
        ---------------     --------------------------------------------------------------------
        **kwargs            Optional additional arguments. See ``Item.clone_items()`` for the full
                            list.
        ===============     ====================================================================

        :return:
            The item corresponding to the cloned experience in the target GIS.
        """

        def _clone_dict(data_dict, source, target, owner, **kwargs):
            """
            Helper function to clone items and update appropriate dict
            """
            new_dict = data_dict
            new_dict["attributes"]["portalUrl"] = target.url
            for k, v in new_dict["dataSources"].items():
                if "itemId" not in v:
                    continue
                v["portalUrl"] = target.url
                item = source.content.get(v["itemId"])
                clone_result = target.content.clone_items([item], owner=owner, **kwargs)
                if clone_result:
                    v["itemId"] = clone_result[0].itemid
                    if "url" in v:
                        v["url"] = clone_result[0].url
                else:
                    targ_item = target.content.search(item.title)[0]
                    v["itemId"] = targ_item.itemid
                    if "url" in v:
                        v["url"] = targ_item.url

            return new_dict

        exp_clone = target.content.clone_items([self._item], owner=owner, **kwargs)
        if exp_clone:
            new_dict = _clone_dict(self._expdict, self._gis, target, owner, **kwargs)
            target_exp = WebExperience(exp_clone[0], gis=target)
            target_exp._expdict = new_dict
            # Create a temporary file and write data to it
            with tempfile.NamedTemporaryFile(
                mode="w+", suffix=".json", delete=False
            ) as tfile:
                json.dump(self._expdict, tfile)
                # Close the file explicitly
                tfile.close()
            target_exp._item.resources.update(
                folder_name="config", file_name="config.json", file=tfile.name
            )
            keywords = target_exp._item.typeKeywords
            for word in keywords:
                if "status" in word:
                    if "Published" in word or "Changed" in word:
                        new_data = _clone_dict(
                            self._item.get_data(), self._gis, target, owner, **kwargs
                        )
                        target_exp._item.update(item_properties={}, data=new_data)
                    else:
                        target_exp._item.update(
                            item_properties={}, data={"__not_publish": True}
                        )
                    break
            return target_exp._item
        else:
            return False



import os
from uuid import uuid4
import json
import traceback
import tempfile
import logging

log = logging.getLogger()

from arcgis.widgets._mapview._raster.engine import RasterData, MatplotlibRasterEngine


class ArcPyRasterEngine(MatplotlibRasterEngine):
    def try_get_raster_data(self, image, bands) -> RasterData:
        try:
            return self._try_get_raster_data(image, bands)
        except Exception as e:
            return super().try_get_raster_data(image, bands)

    def _try_get_raster_data(self, image, bands) -> RasterData:
        from arcpy import Raster, RasterToNumPyArray

        raster = Raster(image)
        img_data = RasterToNumPyArray(raster)
        if len(img_data.shape) == 3:
            img_data = img_data.transpose((1, 2, 0)).squeeze()
            if bands:
                img_data = img_data[:, :, bands].squeeze()
        extent = self._get_extent(raster)
        return RasterData(
            img_data=img_data,
            extent=extent,
            id_=str(hash(raster.extent.JSON + raster.name)),
        )

    def _get_extent(self, raster):
        import arcpy
        from arcpy import Raster, RasterToNumPyArray, SpatialReference

        try:
            transformations = arcpy.ListTransformations(
                raster.spatialReference, arcpy.SpatialReference(4326)
            )
            if len(transformations) > 0:
                transformation = transformations[0]
            else:
                transformation = None
            extent = json.loads(
                raster.extent.projectAs(SpatialReference(4326), transformation).JSON
            )
            if extent["spatialReference"]["wkid"] is not None:
                return extent
            else:
                raise Exception(f"{extent} contains invalid spatial reference")
        except Exception as e:
            log.debug("Handled exception on inferring extent w/ arcpy engine")
            log.debug(e)
            return None



from __future__ import annotations
import sys

import concurrent.futures
from arcgis.gis import GIS, Item
from arcgis.gis.nb import NotebookManager
from arcgis.gis.agonb import AGOLNotebookManager, snapshot as _agosnapshot
from arcgis.gis.nb import _snapshot as _entsnapshot


def create_snapshot(
    item: Item,
    name: str,
    *,
    description: str | None = None,
    notebook_json: dict | None = None,
    access: bool = False,
    server_index: int | None = None,
) -> dict | _entsnapshot.SnapShot | _agosnapshot.SnapShot:
    """
    Creates a Snapshot of a Given Item.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    item                   Required Item. The 'Notebook' typed item to create a snapshot for.
    ------------------     --------------------------------------------------------------------
    name                   Required String.  The name of the snapshot. This is the identifier
                           used to identify the snapshot.
    ------------------     --------------------------------------------------------------------
    description            Optional String. An piece of text that describes the snapshot.
    ------------------     --------------------------------------------------------------------
    notebook_json          Optional Dict. If you want to store different JSON text other
                           than what is in the current notebook provide it here.
    ------------------     --------------------------------------------------------------------
    access                 Optional Bool. When false, the snapshot will not be publicly available.
    ------------------     --------------------------------------------------------------------
    server_index           Optional Int. The ArcGIS Notebook server to use to run the notebook.  This only applies to ArcGIS Enterprise.
    ==================     ====================================================================

    :return: Snapshot | dict (on error)

    """
    gis = item._gis

    mgrs = gis.notebook_server
    if len(mgrs) == 0:
        raise Exception(
            "The user or organization does not have a notebook server configured."
        )
    if gis._portal.is_arcgisonline:
        mgr = mgrs[0]
        sm = mgr.snaphots
        res = sm.create(
            item=item,
            name=name,
            description=description,
            notebook_json=notebook_json,
            access=access,
        )

    else:
        mgr = mgrs[server_index or 0]
        sm = mgr.notebooks.snapshots
        res = sm.create(
            item=item,
            name=name,
            description=description,
            notebook_json=notebook_json,
            access=access,
        )
    if "resourceKey" in res:
        l = [
            s
            for s in item.snapshots
            if s.properties["resourceKey"] == res["resourceKey"]
        ]
        if len(l) > 0:
            return l[0]

    elif "snapshotResourceKey" in res:
        l = [
            s
            for s in item.snapshots
            if s.properties["resourceKey"] == res["snapshotResourceKey"]
        ]
        if len(l) > 0:
            return l[0]

    else:
        return res


def list_snapshots(item: Item) -> list:
    """
    Returns all snapshots associated with a given item.

    :returns: list
    """
    return item.snapshots



from __future__ import annotations
import json
import csv
from datetime import datetime
from arcgis.gis.kubernetes._admin._base import _BaseKube
from collections import OrderedDict
from urllib.request import HTTPError
from arcgis.gis._impl._con import Connection
from arcgis.gis import GIS, Item
from arcgis._impl.common._mixins import PropertyMap
from arcgis._impl.common._isd import InsensitiveDict
from typing import Dict, Any, List, Tuple


class DataStore(_BaseKube):
    _parent = None

    # ----------------------------------------------------------------------
    def __init__(self, url, gis, parent, initialize=False):
        """Constructor


        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        url                    Required string. The machine URL.
        ------------------     --------------------------------------------------------------------
        gis                    Required GIS. The :class:`~arcgis.gis.GIS` object.
        ------------------     --------------------------------------------------------------------
        parent                 Required :class:`~arcgis.gis.admin.kubernetes.DataStores`.
                               The Kubernetes datastore manager class.
        ==================     ====================================================================

        """
        super(DataStore, self).__init__(url=url, gis=gis, parent=parent)
        self._url = url
        self._gis = gis
        self._con = gis._con
        self._parent = parent
        if initialize:
            self._init(gis._con)

    # ----------------------------------------------------------------------
    def delete(self, force=True):
        """
        Removes the datastore from the Kubernetes Site

        :return: Boolean

        """
        if self.properties.systemManaged == False:
            return self._parent._unregister_data_item(
                path=self.properties.path, force=force
            )
        else:
            raise Exception("System Managed DataStore cannot be removed.")
        return False

    # ----------------------------------------------------------------------
    @property
    def status(self):
        """
        This resource returns health information for a **relational data store** only.

        :return: Dict
        """
        url = self._url + "/status"
        params = {"f": "json"}
        return self._con.get(url, params)

    # ----------------------------------------------------------------------
    def switch_role(self):
        """
        This operation promotes a standby relational data store to act as
        primary while also downgrading the existing primary to act as
        standby. This operation may take some time to complete.

        :return: Boolean
        """
        url = self._url + "/switchRole"
        params = {"f": "json"}
        res = self._con.post(url, params)
        if "status" in res:
            return res["status"] == "success"
        return res


###########################################################################
class DataStores(_BaseKube):
    _con = None
    _url = None
    _json_dict = None
    _json = None
    _properties = None

    # ----------------------------------------------------------------------
    def __init__(self, url, gis, initialize=False):
        """Constructor


        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        url                    Required string. The machine URL.
        ------------------     --------------------------------------------------------------------
        gis                    Optional string. The GIS or Server object..
        ==================     ====================================================================

        """
        super(DataStores, self).__init__(gis=gis, url=url)
        self._url = url
        self._gis = gis
        self._con = gis._con
        if initialize:
            self._init(gis._con)

    # ----------------------------------------------------------------------
    def _init(self, connection=None):
        """loads the properties into the class"""
        if connection is None:
            connection = self._con
        params = {"f": "json"}
        try:
            result = connection.get(path=self._url, params=params)
            if isinstance(result, dict):
                self._json_dict = result
                self._properties = InsensitiveDict(result)
            else:
                self._json_dict = {}
                self._properties = InsensitiveDict({})
        except HTTPError as err:
            raise RuntimeError(err)
        except:
            self._json_dict = {}
            self._properties = PropertyMap({})

    # ----------------------------------------------------------------------
    def __str__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """
        returns the object properties
        """
        if self._properties is None:
            self._init()
        return self._properties

    # ----------------------------------------------------------------------
    def __getattr__(self, name):
        """adds dot notation to any class"""
        if self._properties is None:
            self._init()
        try:
            return self._properties.__getitem__(name)
        except:
            for k, v in self._json_dict.items():
                if k.lower() == name.lower():
                    return v
            raise AttributeError(
                "'%s' object has no attribute '%s'" % (type(self).__name__, name)
            )

    # ----------------------------------------------------------------------
    def __getitem__(self, key):
        """helps make object function like a dictionary object"""
        try:
            return self._properties.__getitem__(key)
        except KeyError:
            for k, v in self._json_dict.items():
                if k.lower() == key.lower():
                    return v
            raise AttributeError(
                "'%s' object has no attribute '%s'" % (type(self).__name__, key)
            )
        except:
            raise AttributeError(
                "'%s' object has no attribute '%s'" % (type(self).__name__, key)
            )

    # ----------------------------------------------------------------------
    @property
    def url(self):
        """gets/sets the service url"""
        return self._url

    # ----------------------------------------------------------------------
    @url.setter
    def url(self, value):
        """gets/sets the service url"""
        self._url = value
        self.refresh()

    # ----------------------------------------------------------------------
    def __iter__(self):
        """creates iterable for classes properties"""
        for k, v in self._json_dict.items():
            yield k, v

    # ----------------------------------------------------------------------
    def _refresh(self):
        """reloads all the properties of a given service"""
        self._init()

    # ----------------------------------------------------------------------
    @property
    def stores(self):
        """
        returns a list of all datastores in the enterprise
        """
        stores = []
        for ds in self.properties["items"]:
            url = f"{self._url}/{ds['id']}"
            stores.append(DataStore(url, gis=self._gis, parent=self))
        return stores

    # ----------------------------------------------------------------------
    def add(
        self,
        item: str | Item,
        options: dict[str, Any] | None = None,
        sync: bool | None = None,
    ) -> dict[str, Any] | None:
        """
        Registers a new data item with the data store.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required string. The dictionary representing the data item.
                               See http://resources.arcgis.com/en/help/arcgis-rest-api/index.html#//02r3000001s9000000
        ==================     ====================================================================


        :return:
            The data item if registered successfully, None otherwise.

        """
        if isinstance(item, Item):
            item = item.id
        res = self._register_data_item(item=item, options=options, sync=sync)
        if res["status"] == "success" or res["status"] == "exists":
            url = self._url + f"/{res['id']}"
            return DataStore(url, self._gis, self)
        elif "jobsUrl" in res:
            from arcgis.gis.kubernetes._admin._jobs import Job

            return Job(url=res["JobsUrl"], gis=self._gis)
        else:
            return res

    def validate(self, item: Dict[str, Any]) -> bool:
        """
        Validates that the path (for file shares) or connection string (for
        databases) for a specific data item is accessible to every server
        node in the site by checking against the JSON representing the data
        item, ensuring that the data item can be registered and used
        successfully within the server's data store.

        Validating a data item does not automatically register it for you.
        You need to explicitly register your data item by invoking the
        register operation.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required string. The JSON representing the data item.
                               See http://resources.arcgis.com/en/help/arcgis-rest-api/index.html#//02r3000001s9000000
        ==================     ====================================================================
        """
        params = {"item": item, "f": "json"}
        url = self._url + "/validateDataItem"
        return (
            self._con.post(path=url, postdata=params).get("status", "failed")
            == "success"
        )

    # ----------------------------------------------------------------------
    def _register_data_item(
        self, item: Item, options: dict[str, Any] = None, sync: bool = None
    ):
        """
        Registers a new data item with the server's data store.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required string. The JSON representing the data item.
                               See http://resources.arcgis.com/en/help/arcgis-rest-api/index.html#//02r3000001s9000000
        ==================     ====================================================================

        :return:
            A response
        """
        params = {"item": item, "f": "json"}
        if options:
            params["options"] = options
        if sync:
            params["async"] = sync
        url = self._url + "/registerItem"
        return self._con.post(path=url, postdata=params)

    # ----------------------------------------------------------------------
    def _unregister_data_item(self, path):
        """
        Unregisters a data item that has been previously registered with
        the server's data store.


        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        path                   Required string. The path to the share folder.
        ==================     ====================================================================

        :return:
            Bool

            .. code-block:: python

            EXAMPLE:

            path = r"/fileShares/folder_share"
            print data.unregisterDataItem(path)

        """
        url = self._url + "/unregisterItem"
        params = {"f": "json", "itempath": path, "force": True}
        res = self._con.post(path=url, postdata=params)
        if "status" in res:
            return res["status"] == "success"
        elif "success" in res:
            return res["success"]
        return res

    # ----------------------------------------------------------------------
    def search(
        self,
        parent_path=None,
        ancestor_path=None,
        types=None,
        id=None,
        is_managed=None,
        json=False,
        **kwargs,
    ):
        """
        Use this operation to search through the various data items that are registered in the server's data store.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        parent_path            Optional string. The path of the parent under which to find items.
        ------------------     --------------------------------------------------------------------
        ancestor_path          Optional string. The path of the ancestor under which to find items.
        ------------------     --------------------------------------------------------------------
        types                  Optional string. A filter for the type of the items (for example, fgdb or folder or egdb).
        ------------------     --------------------------------------------------------------------
        id                     Optional string. A filter to search by the ID of the item.
        ------------------     --------------------------------------------------------------------
        is_managed             Optional Boolean.  Specifies if the data store is system managed.
        ------------------     --------------------------------------------------------------------
        json                   Optional Boolean. If `True`, the results will be returned as the raw
                               JSON response. `False`, the response will be a list of `DataStore`
                               objects.  The default is `False`.
        ==================     ====================================================================


        :return:
            A list of the items found matching the search criteria.

        """

        """ jenn note: list of possible types """
        params = {
            "f": "json",
        }
        if parent_path is not None:
            params["parentPath"] = parent_path
        if ancestor_path is not None:
            params["ancestorPath"] = ancestor_path
        if types is not None:
            params["types"] = types
        if id is not None:
            params["id"] = id
        if "decrypt" in kwargs.keys():
            params["decrypt"] = kwargs["decrypt"]
        if is_managed is not None:
            params["isManaged"] = is_managed
        url = self._url + "/findItems"
        res = self._con.post(path=url, postdata=params)
        if "items" in res and json == False:
            stores = []
            for ds in res["items"]:
                url = f"{self._url}/{ds['id']}"
                stores.append(DataStore(url, gis=self._gis, parent=self))
            return stores
        elif "items" in res and json:
            return res["items"]
        return res

    # ----------------------------------------------------------------------
    @property
    def config(self):
        """
        Gets/Sets information on the data store's configuration properties
        that affect the behavior of the data holdings of the server.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        config                 Required string. A JSON string containing the data store configuration.
        ==================     ====================================================================


        :return: dict
        """
        url = self._url + "/config"
        params = {"f": "json"}
        return self._con.get(url, params)

    # ----------------------------------------------------------------------
    @config.setter
    def config(self, config):
        """
        Gets/Sets information on the data store's configuration properties
        that affect the behavior of the data holdings of the server.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        config                 Required string. A JSON string containing the data store configuration.
        ==================     ====================================================================

        :return:
           Dict

        """
        if config is None:
            config = {}
        params = {"f": "json", "datastoreConfig": config}
        url = self._url + "/config/update"
        return self._con.post(path=url, postdata=params)



import importlib.util
from typing import Union, Optional

from arcgis.features.geo import _is_geoenabled
from arcgis.env import active_gis
from arcgis.geometry import find_transformation, SpatialReference, Point
from arcgis.gis import GIS
import pandas as pd
from pandas.api.types import is_numeric_dtype
import numpy as np

arcpy_avail = True if importlib.util.find_spec("arcpy") else False

if arcpy_avail:
    import arcpy


def change_spatial_reference(
    input_dataframe: pd.DataFrame,
    output_spatial_reference: Union[int, SpatialReference] = 4326,
    input_spatial_reference: Optional[Union[int, SpatialReference]] = None,
    transformation_name: Optional[str] = None,
) -> pd.DataFrame:
    """
    Change the spatial reference of the input Spatially Enabled Dataframe to a desired output spatial reference,
        applying a transformation if needed if to the geographic coordinate system changes.
    Args:
        input_dataframe: Valid Spatially Enabled DataFrame
        output_spatial_reference: Optional - Desired output Spatial Reference. Default is
            4326 (WGS84).
        input_spatial_reference: Optional - Only necessary if the Spatial Reference is not
            properly defined for the input data geometry.
        transformation_name: Optional - Transformation name to be used, if needed, to
            convert between spatial references. If not explicitly provided, this will be
            inferred based on the spatial reference of the input data and desired output
            spatial reference.
    Returns: Spatially Enabled DataFrame in the desired output spatial reference.
    """

    # ensure the input spatially enabled dataframe is valid
    assert _is_geoenabled(input_dataframe), "The DataFrame does not appear to be valid."

    # if a spatial reference is set for the dataframe, just use it
    if input_dataframe.spatial.sr is not None:
        in_sr = input_dataframe.spatial.sr

    # if a spatial reference is explicitly provided, but the data does not have one set, use the one provided
    elif input_spatial_reference is not None:
        # check the input
        assert isinstance(input_spatial_reference, int) or isinstance(
            input_spatial_reference, SpatialReference
        ), (
            f"input_spatial_reference must be either an int referencing a wkid or a SpatialReference object, "
            f"not {type(input_spatial_reference)}."
        )

        if isinstance(input_spatial_reference, int):
            in_sr = SpatialReference(input_spatial_reference)
        else:
            in_sr = input_spatial_reference

    # if the spatial reference is not set, common for data coming from geojson, check if values are in lat/lon
    # range, and if so, go with WGS84, as this is likely the case if in this range
    else:
        # get the bounding values for the data
        x_min, y_min, x_max, y_max = input_dataframe.spatial.full_extent

        # check the range of the values, if in longitude and latitude range
        wgs_range = (
            True
            if (x_min > -181 and y_min > -91 and x_max < 181 and y_max < 91)
            else False
        )
        assert wgs_range, (
            "Input data for changing the spatial reference must have a spatial reference set, or one "
            "must be provided."
        )

        # if the values are in range, run with it
        in_sr = SpatialReference(4326)

    # ensure the output spatial reference is a SpatialReference object instance
    if isinstance(output_spatial_reference, SpatialReference):
        out_sr = output_spatial_reference
    else:
        out_sr = SpatialReference(output_spatial_reference)

    # copy the input spatially enabled dataframe since the project function changes the dataframe in place
    out_df = input_dataframe.copy()
    out_df.spatial.set_geometry(input_dataframe.spatial.name)

    # if a transformation was not explicitly provided, see if one is needed
    if transformation_name is None:
        # variable for saving the transformations if needed
        trns_lst = []

        # if arcpy is available, use it to find the transformation
        if arcpy_avail:
            # get any necessary transformations using arcpy, which returns only a list of transformation names
            trns_lst = arcpy.ListTransformations(in_sr.as_arcpy, out_sr.as_arcpy)

        # otherwise we will have to use the geometry rest endpoint to find transformations
        elif transformation_name is None:
            # explicitly ensure find_transformations has a gis instance
            gis = active_gis if active_gis else GIS()

            # get any transformations, if needed due to changing geographic spatial reference, as a list of dicts
            trns_lst = find_transformation(in_sr, out_sr, gis=gis)["transformations"]

        # if a transformation was not explicitly provided and one was discovered to be needed above, get it
        if len(trns_lst) > 0:
            transformation_name = trns_lst[0]

    # project to new spatial reference using the apply method since the geoaccessor project method is not working
    # reliably and only if necessary if the spatial reference is being changed
    if in_sr.wkid != out_sr.wkid:
        out_df[input_dataframe.spatial.name] = out_df[
            input_dataframe.spatial.name
        ].apply(
            lambda geom: geom.project_as(
                out_sr, transformation_name=transformation_name
            )
        )

    # ensure the spatial column is set
    if not _is_geoenabled(out_df):
        out_df.spatial.set_geometry(input_dataframe.spatial.name)

    return out_df


def _get_weighted_centroid_geometry(
    sub_df: pd.DataFrame,
    weighting_column: str,
    sptl_ref: SpatialReference,
    geom_col: str = "SHAPE",
) -> Point:
    """
    Helper function to calculate centroid coordinates.

    Args:
        sub_df: DataFrame to calculate weighted coordinates for.
        weighting_column: Column to be used for weighting.
        sptl_ref: Spatial reference for the output geometry.
        geom_col: Optional - Geometry column, defaults to 'SHAPE'.

    Returns: Tuple of weighted centroid Geometry objects.

    """
    # check if the weighting sum will be naught
    wgt_sum = sub_df[weighting_column].sum()

    # if there is a weighting sum, use it
    if wgt_sum > 0:
        wgt_x = np.average(
            sub_df[geom_col].apply(lambda geom: geom.centroid[0]),
            weights=sub_df[weighting_column],
        )
        wgt_y = np.average(
            sub_df[geom_col].apply(lambda geom: geom.centroid[1]),
            weights=sub_df[weighting_column],
        )

    # if there is not a weighting sum, just get the average centroid
    else:
        wgt_x = np.average(sub_df[geom_col].apply(lambda geom: geom.centroid[0]))
        wgt_y = np.average(sub_df[geom_col].apply(lambda geom: geom.centroid[1]))

    # create a point geometry at the weighted centroid
    wgt_geom = Point({"x": wgt_x, "y": wgt_y, "spatialReference": sptl_ref})

    return wgt_geom


def get_weighted_centroid(
    input_dataframe: pd.DataFrame, grouping_column: str, weighting_column: str
) -> pd.DataFrame:
    """
    Get a spatially enabled dataframe of weighted centroids identified by a grouping
        column.

    Args:
        input_dataframe: Spatially enabled DataFrame with a column identifying unique
            features to be grouped together.
        grouping_column: Column with values for grouping.
        weighting_column: Column with scalar value for weighting.

    Returns: Spatially enabled DataFrame of centroids with the index as values from
        the grouping column.

    """
    # check the input dataframe
    msg_valdf = "A valid Spatially Enabled DataFrame must be provided."
    assert isinstance(input_dataframe, pd.DataFrame), msg_valdf
    assert _is_geoenabled(input_dataframe), msg_valdf

    # ensure the input columns are in the dataframe
    in_cols = input_dataframe.columns
    for col in [grouping_column, weighting_column]:
        msg_col = f'{col} does not appear to be in the DataFrame columns [{", ".join(in_cols)}]'
        assert col in in_cols, msg_col

    # ensure the weighting column is a numeric column
    assert is_numeric_dtype(input_dataframe[weighting_column]), (
        f"The weighting column ({weighting_column}) must be "
        f"numeric, and it appears to be "
        f"{input_dataframe[weighting_column].dtype()}."
    )

    # get the geometry column name
    geom_col = [
        c for c in in_cols if input_dataframe[c].dtype.name.lower() == "geometry"
    ][0]

    # get the input spatial reference
    sptl_ref = input_dataframe.spatial.sr

    # calculate weighted centroids
    centroid_df = (
        input_dataframe.groupby(grouping_column)
        .apply(
            lambda sub_df: _get_weighted_centroid_geometry(
                sub_df, weighting_column, sptl_ref, geom_col
            )
        )
        .to_frame("SHAPE")
    )

    # reset the geometry to ensure updates recognized
    centroid_df.spatial.set_geometry("SHAPE")

    return centroid_df



from __future__ import annotations
from typing import TypeVar
from arcgis.gis import GIS, Item
from arcgis._impl.common._mixins import PropertyMap
import concurrent.futures

__all__ = ["NotebookManager"]

K = TypeVar("K")
V = TypeVar("V")


########################################################################
class NotebookManager(object):
    """
    Provides access to managing a site's notebooks. An object of this
    class can be created using :attr:`~arcgis.gis.nb.NotebookServer.notebooks` property of the
    :class:`~arcgis.gis.nb.NotebookServer` class
    """

    _url = None
    _gis = None
    _properties = None
    _nbs = None
    _snapshot = None

    # ----------------------------------------------------------------------
    def __init__(self, url, gis, nbs):
        """Constructor"""
        self._url = url
        self._nbs = nbs
        if isinstance(gis, GIS):
            self._gis = gis
            self._con = self._gis._con
        else:
            raise ValueError("Invalid GIS object")

    # ----------------------------------------------------------------------
    def _init(self):
        """loads the properties"""
        try:
            params = {"f": "json"}
            res = self._gis._con.get(self._url, params)
            self._properties = PropertyMap(res)
        except:
            self._properties = PropertyMap({})

    # ----------------------------------------------------------------------
    def __str__(self):
        return "< NotebookManager @ {url} >".format(url=self._url)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "< NotebookManager @ {url} >".format(url=self._url)

    # ----------------------------------------------------------------------
    @staticmethod
    def _future_job(
        fn,
        task_name,
        jobid=None,
        task_url=None,
        notify=False,
        gis=None,
        **kwargs,
    ):
        """
        runs the job asynchronously

        :return: Job object
        """
        from arcgis._impl._async.jobs import Job

        tp = concurrent.futures.ThreadPoolExecutor(1)
        try:
            future = tp.submit(fn=fn, **kwargs)
        except:
            future = tp.submit(fn, **kwargs)
        tp.shutdown(False)
        return Job(future, task_name, jobid, task_url, notify, gis=gis)

    # ----------------------------------------------------------------------
    def execute_notebook(
        self,
        item: Item,
        update_portal_item: bool = True,
        parameters: list | None = None,
        save_parameters: bool = False,
        instance_type: str | None = None,
        timeout: float | int = 50,
        future: bool = False,
    ):
        """

        The Execute Notebook operation allows administrators to remotely
        run a notebook in their ArcGIS Notebook Server site. The notebook
        specified in the operation will be run with all cells in order.

        Using this operation, you can schedule the execution of a notebook,
        either once or with a regular occurrence. This allows you to
        automate repeating tasks such as data collection and cleaning,
        content updates, and portal administration. On Linux machines, use
        a cron job to schedule the executeNotebook operation; on Windows
        machines, you can use the Task Scheduler app.

        .. note::
            To run this operation, you must be logged in with an ArcGIS
            Enterprise portal account. You cannot execute notebooks from
            the ArcGIS Notebook Server primary site administrator
            account.

        You can specify parameters to be used in the notebook at execution
        time. If you've specified one or more parameters, they'll be
        inserted into the notebook as a new cell. This cell will be placed
        at the beginning of the notebook, unless you have added the tag
        parameters to a cell.

        ====================    ====================================================================
        **Parameter**            **Description**
        --------------------    --------------------------------------------------------------------
        item                    Required :class:`~arcgis.gis.Item`. Opens an existing portal item.
        --------------------    --------------------------------------------------------------------
        update_portal_item      Optional Boolean. Specifies whether you want to update the
                                notebook's portal item after execution. The default is true. You may
                                want to specify true when the notebook you're executing contains
                                information that needs to be updated, such as a workflow that
                                collects the most recent version of a dataset. It may not be
                                important to update the portal item if the notebook won't store any
                                new information after executing, such as an administrative notebook
                                that emails reminders to inactive users.
        --------------------    --------------------------------------------------------------------
        parameters              Optional List. An optional array of parameters to add to the
                                notebook for this execution. The parameters will be inserted as a
                                new cell directly after the cell you have tagged ``parameters``.
                                Separate parameters with a comma. Use the format "x":1 when
                                defining parameters with numbers, and "y":"text" when defining
                                parameters with text strings.
        --------------------    --------------------------------------------------------------------
        save_parameters         Optional Boolean.  Specifies whether the notebook parameters cell
                                should be saved in the notebook for future use. The default is
                                false.
        --------------------    --------------------------------------------------------------------
        instance_type           Optional String. The instance type.
        --------------------    --------------------------------------------------------------------
        timeout                 Optional Int. The number of minutes to run the instance before timeout.
        --------------------    --------------------------------------------------------------------
        future                  Optional boolean. If True, a Job object will be returned and the process
                                will not wait for the task to complete. The default is False, which means wait for results.
        ====================    ====================================================================

        :return: Dict else If ``future = True``, then the result is
                 a `concurrent.futures.Future <https://docs.python.org/3/library/concurrent.futures.html>`_ object.
                 Call ``result()`` to get the response

        """
        from arcgis.gis import Item

        url = self._url + "/executeNotebook"
        itemid = None
        if isinstance(item, str):
            itemid = item
        elif isinstance(item, Item):
            itemid = item.itemid
        params = {
            "f": "json",
            "itemId": itemid,
            "updatePortalItem": update_portal_item,
            "saveInjectedParameters": save_parameters,
            "instanceTypeName": instance_type,
            "executionTimeoutInMinutes": timeout,
        }
        if parameters:
            params["notebookParameters"] = parameters
        if future:

            def _fn(url, params, nbs):
                import time

                start_job = self._gis._con.post(url, params)
                if "jobUrl" in start_job:
                    resp = self._gis._con.get(start_job["jobUrl"], {"f": "json"})
                else:
                    return start_job
                if "status" in resp and resp["status"].lower() != "success":
                    status = self._gis._con.get(start_job["jobUrl"], {"f": "json"})
                    i = 0
                    while status["status"].lower() != "completed":
                        time.sleep(0.3 * i)
                        if status["status"].lower() == "failed":
                            return status
                        elif (
                            status["status"].lower().find("fail") > -1
                            or status["status"].lower().find("error") > -1
                        ):
                            raise Exception(f"Job Fail {status}")
                        status = self._gis._con.get(start_job["jobUrl"], {"f": "json"})
                        i += 1
                        if i > 20:
                            i = 20
                    return status
                return resp

            return NotebookManager._future_job(
                fn=_fn,
                task_name="Execute Notebook",
                gis=self._gis,
                **{"url": url, "params": params, "nbs": self._nbs},
            )
        res = self._gis._con.post(url, params)
        return res

    # ----------------------------------------------------------------------
    def open_notebook(
        self,
        itemid: str,
        templateid: str | None = None,
        nb_runtimeid: str | None = None,
        template_nb: str | None = None,
        instance_type: str | None = None,
        *,
        future: bool = False,
    ):
        """

        Opens a notebook on the notebook server

        ==================      ====================================================================
        **Parameter**            **Description**
        ------------------      --------------------------------------------------------------------
        itemid                  Required String. Opens an existing portal item.
        ------------------      --------------------------------------------------------------------
        templateid              Optional String. The id of the portal notebook template. To get the
                                system templates, look at the sample notebooks group:

                                .. code-block:: python

                                    >>> from arcgis.gis import GIS
                                    >>> gis = GIS()
                                    >>> grp = gis.groups.search("title:(esri sample notebooks) AND
                                    >>>                                 owner:\"esri_notebook\")[0]
                                    >>> grp.content
        ------------------      --------------------------------------------------------------------
        nb_runtimeid            Optional String. The runtime to use to generate a new notebook.
        ------------------      --------------------------------------------------------------------
        template_nb             Optional String. The start up template for the notebook.
        ------------------      --------------------------------------------------------------------
        instance_type           Optional String. The name of the instance type.
        ------------------      --------------------------------------------------------------------
        future                  Optional Bool.
        ==================      ====================================================================

        :return: Dict

        """

        def _fn(url, params, nbs):
            """used to fire off async job"""
            import time

            start_job = self._gis._con.post(url, params)
            status_url = start_job.get("jobUrl", None) or start_job.get(
                "notebookStatusUrl", None
            )
            if status_url:
                resp = self._gis._con.get(status_url, {"f": "json"})
            else:
                return start_job
            if "status" in resp and resp["status"].lower() != "success":
                status = self._gis._con.get(status_url, {"f": "json"})
                i = 0
                while status["status"].lower() != "completed":
                    time.sleep(0.3 * i)
                    if status["status"].lower() == "failed":
                        return status
                    elif (
                        status["status"].lower().find("fail") > -1
                        or status["status"].lower().find("error") > -1
                    ):
                        raise Exception(f"Job Fail {status}")
                    status = self._gis._con.get(status_url, {"f": "json"})
                    i += 1
                    if i > 20:
                        i = 20
                return status

        if hasattr(itemid, "id"):
            itemid = getattr(itemid, "id")
        params = {
            "itemId": itemid,
            "templateId": templateid,
            "notebookRuntimeId": nb_runtimeid,
            "templateNotebook": template_nb,
            "async": True,
            "f": "json",
            "instanceTypeName": instance_type,
        }
        url = self._url + "/openNotebook"

        if future:
            return NotebookManager._future_job(
                fn=_fn,
                task_name="Open Notebook",
                gis=self._gis,
                **{"url": url, "params": params, "nbs": self._nbs},
            )
        else:
            res = self._con.post(url, params)
            status_url = res.get("jobUrl", None) or res.get("notebookStatusUrl", None)
            if status_url:
                import time

                job_url = status_url
                params = {"f": "json"}
                job_res = self._con.get(job_url, params)
                while job_res["status"].upper() != "COMPLETED":
                    job_res = self._con.get(job_url, params)
                    if job_res["status"].lower().find("fail") > -1:
                        return job_res
                    time.sleep(2.5)
                return job_res
            return res



from __future__ import annotations
import datetime
from typing import Optional
from arcgis import GIS, env
from arcgis._impl.common._mixins import PropertyMap

########################################################################


class TopographicProductionManager(object):
    """
    The Topographic Production Service resource represents a topographic
    production server object extension (SOE). It is the entry point
    for all functionality related to creating topographic products in ArcGIS Server.
    This resource provides information about the service and enables
    the topographic production server capabilities on a map service.
    You can also view child resources and operations defined by the SOE.

    .. note::
        The use of this resource requires an ArcGIS GIS Server
        Advanced license and a Production Mapping Server or
        Defense Mapping Server license.

    =====================   ===========================================
    **Inputs**              **Description**
    ---------------------   -------------------------------------------
    url                     Required String. The web endpoint to the topographic service.
    ---------------------   -------------------------------------------
    gis                     Optional GIS. The enterprise connection to
                            the Portal site. A connection can be passed
                            in such as a Service Directory connection.
    =====================   ===========================================


    """

    _con = None
    _gis = None
    _url = None
    _property = None
    # ----------------------------------------------------------------------

    def __init__(self, url, gis=None):
        """Constructor"""
        if gis is None:
            gis = env.active_gis
        if isinstance(gis, GIS):
            self._gis = gis
            self._con = self._gis._portal.con
        elif hasattr(gis, "_con"):
            self._gis = gis
            self._con = gis._con
        else:
            raise ValueError("gis must be of type GIS")
        self._gis = gis
        self._url = url

    # ----------------------------------------------------------------------
    def _init(self):
        """initializer"""
        try:
            res = self._con.get(self._url, {"f": "json"})
            self._property = PropertyMap(res)
        except Exception as e:
            self._property = PropertyMap({})

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the properties for the service"""
        if self._property is None:
            self._init()
        return self._property

    # ----------------------------------------------------------------------
    def add_product(
        self,
        product_definition: dict,
        raster: str = None,
        ancillary_layers: list[dict] = None,
    ):
        """
        The add_product operation adds a definition of a map product to
        the Topographic Production Service resource that can be used to generate a map

        ======================      =====================================================
        **Parameter**                **Description**
        ----------------------      -----------------------------------------------------
        product_definition          Required dictionary. The definition of the map product.

                                    `See Parameters <https://developers.arcgis.com/rest/services-reference/enterprise/tps-add-product.htm#:~:text=productDefinition%20properties/>`_
        ----------------------      -----------------------------------------------------
        raster                      Optional string. The path to the raster on disk (server path).
        ----------------------      -----------------------------------------------------
        ancillary_layers            Optional list of dictionaries. Additional layers to
                                    include in the final product.

                                    Syntax:
                                    ```
                                    [{
                                       "layer": "url of the layer",
                                        "featureClass": "name of the feature class to extract to",
                                        "map": "name of the map the layer will be inserted into",
                                        "layerIndex": "insertion index of the layer"
                                        },
                                        ...
                                    ]
                                    ```

                                    .. note::
                                        This parameter supports services located in the
                                        same portal site as the server object extension (SOE)
                                        or services that are publicly available.
                                        The featureClass, map, and layerIndex properties
                                        in the array are optional. If the dataset is
                                        identifiable from the feature service, it is not
                                        necessary to provide the featureClass property.
                                        The default values are 0 for layerIndex and BaseMap
                                        for map.
        ======================      =====================================================

        :return: True if succeeded and False if failed.
        """
        url = "%s/addProduct" % self._url
        params = {
            "f": "json",
            "productDefinition": product_definition,
            "raster": raster,
            "ancillaryLayers": ancillary_layers,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def calculate_extent(
        self,
        name: str,
        version: str,
        area_interest_layer: str,
        input_geometry: dict,
        out_sr: str,
    ):
        """
        The calculate_extent operation calculates a custom area of interest (AOI)
        for a given product and version. The result can be specified as the value for the customAoi parameter of the generateProduct operation.

        ======================      =====================================================
        **Parameter**                **Description**
        ----------------------      -----------------------------------------------------
        name                        Required string. The name of the product.
        ----------------------      -----------------------------------------------------
        version                     Required string. The version of the product to generate.
        ----------------------      -----------------------------------------------------
        area_interest_layer         Required string. The url of the layer defining the
                                    product's area of interest.
        ----------------------      -----------------------------------------------------
        input_geometry              Required dictionary. A geometry defining the center
                                    of the custom extent. A point geometry.
        ----------------------      -----------------------------------------------------
        out_sr                      Required string. The coordinate system for the
                                    extent that is returned.
        ======================      =====================================================

        :return: True if success and False otherwise.
        """
        url = "%s/calculateExtent" % self._url
        params = {
            "f": "json",
            "productName": name,
            "productVersion": version,
            "aoiLayer": area_interest_layer,
            "inputGeometry": input_geometry,
            "outSpatialReference": out_sr,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def generate_product(
        self,
        name: str,
        version: str,
        area_interest_layer: str,
        area_interest_feature_id: str,
        output_type: str,
        custom_area_interest: Optional[list[dict]] = None,
        layer_exclusion: Optional[list[str]] = None,
        ancillary_layers: Optional[list[dict]] = None,
    ):
        """
        The generate_product operation automates the process of producing
        a layout or map based on an existing map product definition.


        ========================        =====================================================
        **Parameter**                    **Description**
        ------------------------        -----------------------------------------------------
        name                            Required string. The name of the product.
        ------------------------        -----------------------------------------------------
        version                         Required string. The version of the product to generate.
        ------------------------        -----------------------------------------------------
        area_interest_layer             Required string. The url of the layer defining the
                                        product's area of interest.
        ------------------------        -----------------------------------------------------
        area_interest_feature_id        Required string. The Object ID of the area of interest
                                        feature in the area_interest_layer.

                                        .. note::
                                            If you use the optional custom_area_interest
                                            parameter, then this parameter is ignored.
        ------------------------        -----------------------------------------------------
        output_type                     Required string. The type of output.

                                        `Values: "aprx" | "pagx" | "pdf"`
        ------------------------        -----------------------------------------------------
        custom_area_interest            Optional list dictionary. The features defining the
                                        product's area of interest. One feature is allowed at
                                        the current release.
        ------------------------        -----------------------------------------------------
        layer_exclusion                 Optional list of strings. The list of layer names to
                                        exclude from the product.
        ------------------------        -----------------------------------------------------
        ancillary_layers                Optional list of dictionaries. Additional layers to
                                        include in the final product.

                                        Syntax:
                                        ```
                                        [{
                                        "layer": "url of the layer",
                                            "featureClass": "name of the feature class to extract to",
                                            "map": "name of the map the layer will be inserted into",
                                            "layerIndex": "insertion index of the layer"
                                            },
                                            ...
                                        ]
                                        ```

                                        Ancillary layers are added as feature service layers
                                        of the final product and don't extract data to the
                                        local geodatabase.

                                        .. note::
                                            This parameter supports services located in the
                                            same portal site as the server object extension (SOE)
                                            or services that are publicly available.
                                            The featureClass, map, and layerIndex properties
                                            in the array are optional. If the dataset is
                                            identifiable from the feature service, it is not
                                            necessary to provide the featureClass property.
                                            The default values are 0 for layerIndex and BaseMap
                                            for map.
        ========================        =====================================================


        :return: A dictionary with the jobId and statusUrl.
        """
        url = "%s/generateProduct" % self._url
        params = {
            "f": "json",
            "productName": name,
            "productVersion": version,
            "aoiLayer": area_interest_layer,
            "aoiFeatureId": area_interest_feature_id,
            "outputType": output_type,
            "customAoi": custom_area_interest,
            "layerExclusion": layer_exclusion,
            "ancillaryLayers": ancillary_layers,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def remove_product(self, name: str):
        """
        The remove_product operation removes a product from the Topographic
        Production service and returns a standard REST success or error message.

        ========================        =====================================================
        **Parameter**                    **Description**
        ------------------------        -----------------------------------------------------
        name                            Required string. The name of the product.
        ========================        =====================================================

        :return: True if success, else False.
        """
        url = "%s/removeProduct" % self._url
        params = {
            "f": "json",
            "productName": name,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def update_product(
        self,
        name: str,
        description: Optional[str] = None,
        enabled: Optional[bool] = None,
        sheet_id_field: Optional[str] = None,
        raster: Optional[str] = None,
        versions: Optional[dict] = None,
        ancillary_layers: Optional[dict] = None,
    ):
        """
        The update_product operation updates properties of a product.

        ========================        =====================================================
        **Parameter**                    **Description**
        ------------------------        -----------------------------------------------------
        name                            Required string. The name of the product.
        ------------------------        -----------------------------------------------------
        description                     Optional string. The new description of the product.
        ------------------------        -----------------------------------------------------
        enabled                         Optional bool. Enables or disables the product. If a
                                        product is disabled, the `generate_product`
                                        operation returns an error for that product.
        ------------------------        -----------------------------------------------------
        sheet_id_field                  Optional string. Updates the sheet ID field.
        ------------------------        -----------------------------------------------------
        raster                          Optional string. The path to a raster on disk (server path).
        ------------------------        -----------------------------------------------------
        versions                        Optional dict. Indicates the type of operation and
                                        versions. There are two formats.

                                        If the operation is `add` or `update`:
                                        ```
                                        {
                                           "operation": "add | update",
                                            "versions": [
                                                {
                                                    "name":"name of version",
                                                    "template": "name of template"
                                                },
                                                ...
                                            ]
                                        }
                                        ```

                                        If the operation is `remove`:
                                        ```
                                        {
                                            "operation": "remove",
                                            "versions": [
                                                    "name of version 1",
                                                    "name of version 2",
                                                    ...
                                            ]
                                        }
        ------------------------        -----------------------------------------------------
        ancillary_layers                Optional list of dictionaries. Additional layers to
                                        include in the final product.

                                        Syntax:
                                        ```
                                        [{
                                        "layer": "url of the layer",
                                            "featureClass": "name of the feature class to extract to",
                                            "map": "name of the map the layer will be inserted into",
                                            "layerIndex": "insertion index of the layer"
                                            },
                                            ...
                                        ]
                                        ```

                                        Ancillary layers are added as feature service layers
                                        of the final product and don't extract data to the
                                        local geodatabase.

                                        .. note::
                                            This parameter supports services located in the
                                            same portal site as the server object extension (SOE)
                                            or services that are publicly available.
                                            The featureClass, map, and layerIndex properties
                                            in the array are optional. If the dataset is
                                            identifiable from the feature service, it is not
                                            necessary to provide the featureClass property.
                                            The default values are 0 for layerIndex and BaseMap
                                            for map.
        ========================        =====================================================

        :return: The product name and whether the operation was a success(True) or a failure (false)
        """
        url = "%s/updateProduct" % self._url
        params = {
            "f": "json",
            "productName": name,
            "description": description,
            "enabled": enabled,
            "sheetIDField": sheet_id_field,
            "raster": raster,
            "productVersions": versions,
            "ancillaryLayers": ancillary_layers,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def products(self, include_def: bool = True):
        """
        The products operation retrieves the products that a Topographic
        Production Service resource supports.

        ====================    ==========================================
        **Parameter**            **Description**
        --------------------    ------------------------------------------
        include_def             Optional bool. Specifies whether the full
                                json definition of the map product is
                                included.
        ====================    ==========================================

        :return: The products that are supported as a dictionary.
        """
        url = "%s/products" % self._url
        params = {"f": "json", "includeDef": include_def}
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def product(self, name: str, include_def: bool = True):
        """
        The product operation retrieves a single product from the products that
        a Topographic Production Service resource supports.

        ====================    ==========================================
        **Parameter**            **Description**
        --------------------    ------------------------------------------
        name                    Required string. The name of the product.
        --------------------    ------------------------------------------
        include_def             Optional bool. Specifies whether the full
                                json definition of the map product is
                                included.
        ====================    ==========================================
        """
        url = "%s/products/%s" % (self._url, name)
        params = {"f": "json", "includeDef": include_def}
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def jobs_manager(self):
        """
        Retrieve the Topographic Production Job Manager class. With this manager
        you can retrieve all jobs, a single job, query or cancel a job.

        :return:
           :class:`~arcgis.features._topographic.TopographicProductionJobManager`
        """
        url = "%s/jobs" % self._url
        return TopographicProductionJobManager(url, self._gis)


# ============================================================================
class TopographicProductionJobManager(object):
    def __init__(self, url: str, gis: GIS):
        self._url = url
        self._gis = gis
        self._con = gis._con

    # ----------------------------------------------------------------------
    def jobs(self):
        """
        Retrieve all the jobs for the service.
        """
        params = {
            "f": "json",
        }
        return self._con.post(self.url, params)

    # ----------------------------------------------------------------------
    def job(self, job_id: str, msg_level: str = None):
        """
        The job operation tracks the status of a job executed by the
        generate_product REST operation and returns the status, start date,
        last modified date, and messages of the job.

        ====================    ==========================================
        **Parameter**            **Description**
        --------------------    ------------------------------------------
        job_id                  Required string. The unique job id to see.
        --------------------    ------------------------------------------
        msg_level               Optional string. The message level associated
                                with the job.

                                `Values: "info" | "warn" | "error"`

                                .. note::
                                    Regardless of the msgLevel value,
                                    any errors that a job contains are
                                    included in the messages array of
                                    the response by default.
        ====================    ==========================================

        :return:
            A Json response with syntax:

            ```
            {
                "status": <untranslated string representing general status of the job>,
                "statusCode": <code representing specific status of the job>,
                "submissionTime": <time and date of the job submission in UTC and ISO 8601 format YYYY-MM-DDThh:mm:ssZ>,
                "lastUpdatedTime": <time and date of the last job update in UTC and ISO 8601 format YYYY-MM-DDThh:mm:ssZ>,
                "percentComplete": <percent of the job completed>,
                "aoiLayer": <URL of AOI layer>,
                "aoiFeatureId": <AOI feature ID>,
                "outputUrl": <URL of output>,
                "user": <username>,
                "jobId": <job identifier>,
                "productName": <name of the product>,
                "productVersion": <version of the product>,
                "outputType": <type/format of the output>,
                "messages": {
                <informative | error | warning>
                }
            }
            ```
        """
        url = "%s/job/%s" % (self._url, job_id)
        params = {"f": "json", "msgLevel": msg_level}
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def cancel_job(self, job_id: str):
        """
        The cancel operation cancels a job submitted through the `generate_product`
        REST operation. It returns a standard REST success or error message.

        ====================    ==========================================
        **Parameter**            **Description**
        --------------------    ------------------------------------------
        job_id                  Required string. The unique job id to cancel.
        ====================    ==========================================

        :return: Success (true) or Failure (false)
        """
        url = "%s/cancel" % self._url
        params = {"f": "json", "jobId": job_id}
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def query_job(
        self,
        status: Optional[int] = None,
        start_date: datetime.datetime = None,
        end_date: datetime.datetime = None,
        msg_level: Optional[str] = None,
    ):
        """
        The query operation retrieves the status of jobs executed by the
        `generate_product` REST operation and returns the status, start date,
        last modified date, and messages for a set of jobs.

        ====================    ==========================================
        **Parameter**            **Description**
        --------------------    ------------------------------------------
        status                  Optional int. Retrieves all the jobs with
                                a particular status.

                                `Values:
                                    * New—0
                                    * Submitted—1
                                    * Waiting—2
                                    * Executing—3
                                    * Succeeded—4
                                    * Failed—5
                                    * TimedOut—6
                                    * Canceling—7
                                    * Canceled—8
                                    * Deleting—9
                                    * Deleted—10`
        --------------------    ------------------------------------------
        start_date              Optional datetime. Retrieves all the jobs
                                that started on or after the start date and
                                time specified. The date and time should be
                                UTC and in ISO 8601 format. If time is not
                                specified, then midnight is used in the
                                following format: 00:00:00.
        --------------------    ------------------------------------------
        end_date                Optional datetime. Retrieves all the jobs
                                that ended on or after the end date and
                                time specified. The date and time should be
                                UTC and in ISO 8601 format. If time is not
                                specified, then 23:59:59 is used.
        --------------------    ------------------------------------------
        msg_level               Optional string. The message level associated
                                with the job.

                                `Values: "info" | "warn" | "error"`

                                .. note::
                                    Regardless of the msgLevel value,
                                    any errors that a job contains are
                                    included in the messages array of
                                    the response by default.
        ====================    ==========================================
        """
        url = "%s/query" % self._url
        params = {
            "f": "json",
            "status": status,
            "startDate": start_date,
            "endDate": end_date,
            "msgLevel": msg_level,
        }
        return self._con.post(url, params)



import json
import uuid
from arcgis.gis import GIS
from arcgis._impl.common._isd import InsensitiveDict
from arcgis.gis._impl._con._url_validator import validate_url
from pathlib import Path
from ._base import BaseOGC


def _is_file(path):
    """checks if the data is a file"""
    return Path(path).is_file()


###########################################################################
class GeoJSONLayer(BaseOGC):
    """
    The GeoJSONLayer class is used to create a layer based on GeoJSON.
    GeoJSON is a format for encoding a variety of geographic data
    structures. The GeoJSON data must comply with the RFC 7946
    specification which states that the coordinates are in
    spatial reference: WGS84 (wkid 4326).


    ===============     ====================================================================
    **Parameter**        **Description**
    ---------------     --------------------------------------------------------------------
    url                 Optional String. The web location of the GeoJSON file.
    ---------------     --------------------------------------------------------------------
    data                Optional String or Dict. A path to a GeoJSON file, the GeoJSON data as a string, or the GeoJSON data as a dictionary.
    ---------------     --------------------------------------------------------------------
    copyright           Optional String. Describes limitations and usage of the data.
    ---------------     --------------------------------------------------------------------
    opacity             Optional Float.  This value can range between 1 and 0, where 0 is 100 percent transparent and 1 is completely opaque.
    ---------------     --------------------------------------------------------------------
    renderer            Optional Dictionary. A custom set of symbology for the given geojson dataset.
    ---------------     --------------------------------------------------------------------
    scale               Optional Tuple. The min/max scale of the layer where the positions are: (min, max) as float values.
    ---------------     --------------------------------------------------------------------
    title               Optional String. The title of the layer used to identify it in places such as the Legend and Layer List widgets.
    ===============     ====================================================================


    """

    _type = "geojson"
    _url = ""
    _data = {}

    # ----------------------------------------------------------------------
    def __init__(self, url=None, data=None, **kwargs):
        """init"""
        super(GeoJSONLayer, self)
        if url is None and data is None:
            raise Exception("A `url` or `data` must be given to proceed")
        if isinstance(data, str) and _is_file(data):
            with open(data, "r") as r:
                self._data = json.loads(r.read())
        elif isinstance(data, str) and _is_file(data) == False:
            self._data = json.loads(data)
        elif isinstance(data, dict):
            self._data = dict(data)
        elif url is None and data and not isinstance(data, (str, dict)):
            raise ValueError("`data` must be of type string or dict.")
        if url and validate_url(url) == False:
            raise ValueError(f"Invalid `url` : {url}")
        self._url = url
        self._type = "GeoJSON"
        self._copyright = kwargs.pop("copyright", "")
        self._title = kwargs.pop("title", "GeoJSON Layer")
        self._id = kwargs.pop("id", uuid.uuid4().hex)  # hidden input, but accepted
        self._min_scale, self._max_scale = kwargs.pop("scale", (0, 0))
        self._opacity = kwargs.pop("opacity", 1)
        if "renderer" in kwargs:
            r = kwargs.pop("renderer", None)
            if isinstance(r, dict):
                self._renderer = InsensitiveDict(r)
            else:
                self._renderer = None
        else:
            self._renderer = None

    # ----------------------------------------------------------------------
    @property
    def renderer(self) -> InsensitiveDict:
        """Gets/Sets the renderer for the layer"""
        return self._renderer

    # ----------------------------------------------------------------------
    @renderer.setter
    def renderer(self, renderer: dict):
        """Gets/Sets the renderer for the layer"""
        if isinstance(renderer, dict) and renderer:
            self._renderer = InsensitiveDict(renderer)

    # ----------------------------------------------------------------------
    @property
    def _lyr_json(self) -> dict:
        """Represents the MapView widget's JSON format"""
        lyr = {
            "type": self._type,
            "url": self._url,
            "data": self._data,
            "copyright": self._copyright,
            "title": self._title,
            "id": self._id,
            "minScale": self.scale[0],
            "maxScale": self.scale[1],
            "opacity": self._opacity,
        }
        if self._renderer:
            lyr["renderer"] = self._renderer._json()
        return lyr

    @property
    def _operational_layer_json(self) -> dict:
        """Represents the WebMap's JSON format"""
        return self._lyr_json

    @property
    def url(self):
        """
        Get/Set the data associated with the GeoJSON Layer

        :return: String

        """
        return self._url or self._data

    @url.setter
    def url(self, data):
        """
        Get/Set the data associated with the GeoJSON Layer

        The data can be a string, file or URL.

        :return: String
        """

        if validate_url(data):
            self._url = data
        elif _is_file(data):
            with open(data, "r") as r:
                self._data = r.read()
        elif isinstance(data, str):
            self._data = data
        else:
            raise ValueError("The data must be a valid URL, file, or text blob.")



"""
This resource is a container for all the KMZ files created on the
server.s
"""
from __future__ import absolute_import
from __future__ import print_function
from .._common import BaseServer
from arcgis.gis import GIS


########################################################################
class KML(BaseServer):
    """
    This resource is a container for all the KMZ files created on the
    server.
    """

    _con = None
    _url = None
    _json_dict = None

    # ----------------------------------------------------------------------
    def __init__(self, url: str, gis: GIS, initialize: bool = False):
        """
        Constructor

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        url                 Required string. The administration URL for the ArcGIS Server.
        ---------------     --------------------------------------------------------------------
        gis                 Required Server object. Connection object.
        ---------------     --------------------------------------------------------------------
        initialize          Optional boolean. If true, information loaded at object
        ===============     ====================================================================

        """
        super(KML, self).__init__(gis=gis, url=url)
        self._con = gis
        self._url = url
        if initialize:
            self._init(gis)

    # ----------------------------------------------------------------------
    def create_KMZ(self, kmz_as_json: dict) -> dict:
        """
        Creates a KMZ file from json.
        See https://developers.arcgis.com/rest/enterprise-administration/server/createkmz.htm
        for more information.
        """
        url = self._url + "/createKmz"
        params = {"f": "json", "kml": kmz_as_json}
        return self._con.post(path=url, postdata=params)
